/*
 * Copyright 2001, Olivier Chapuis. No guarantees or warantees or anything
 * are provided or implied in any way whatsoever. Use this program at your
 * own risk. Permission to use this program for any purpose is given,
 * as long as the copyright is kept intact. 
 */

/* This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* I used here and in the other parts of this module Code from FvwmPager, 
 * FvwmWinList, fvwm/gnome.c, I also take a look at Gnome AfterStep Module.
 * So:
 * 
 * the concept for interfacing this module to the Window Manager, are all 
 * original work by Robert Nation, Copyright 1994, Robert Nation.
 *
 * Thanks to: 
 * Mike Finger (mfinger@mermaid.micro.umn.edu or Mike_Finger@atk.com)
 * Nobutaka Suzuki.
 * Raster and Jay Painter <jpaint@gnu.org>
 * Rafal Wierzbicki <rafal@mcss.mcmaster.ca>
 *
 * Thanks to all FVWM WORKERS! 
 */

#include "config.h"

#include <X11/Intrinsic.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>
#include <X11/Xproto.h>

#include <stdio.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/time.h>
#include <stdlib.h>

#ifdef HAVE_SYS_BSDTYPES_H
#include <sys/bsdtypes.h>
#endif

#include "libs/Module.h"
#include "libs/fvwmsignal.h"

#include "FvwmNetHints.h"
#include "parse.h"
#include "net.h"
#include "windowlist.h"
#include "icon.h"
#include "net_debug.h"
#include "utf8.h"

/* basic */
char *MyName;
fd_set_size_t fd_width;
int fd[2];
int x_fd;
int Clength;
char *ImagePath;

/* X related things */
ScreenInfo Scr;
Display *dpy;
/* our list of windows */
NetWindow *Start = NULL;
NetWindow *FocusWin = NULL;
WindowList Stacking;
WindowList KdeSysTray;
PingWindowList PingList;

#if FNH_IMLIB
ImlibData *ImlibId;
#endif

extern Atom ATOM_NET_WM_WINDOW_TYPE_DESKTOP;

/* config (see also net.c) */
long NumberOfDesktops = 4;   /* not in Scr because this is not an FVWM concept */
unsigned DoSendMiniIcon;
unsigned MiniIconOverrideMode;
unsigned IconOverrideMode;

Bool KsmServerWorkAround;

extern Bool PagerSkipListUseSkipList;
extern int MiniIconSize[2];
extern Bool UseDynamicStrutMaximize;
extern Bool UseStrutPlacement;
extern Bool UseExtendedName;
extern unsigned DefaultHorizMaximize;
extern unsigned DefaultVertMaximize;
extern int StaysOnTopLayer;
extern int StaysPutLayer;

extern char *FnhSysStartStopFile;
extern char *FnhSysDesktopStartStopFile;

/* module */
int Started = 0;
char *NetAnimCommand = NULL;
char *NetAnimWindowName = NULL;
Window NetAnimWindow = None;
Bool NetAnimWindowState = True;
#if HAVE_ICONV
char *IconvCharset = NULL;
#endif

/* alarm for PING action */
RETSIGTYPE Alarm(int sig);
void ClearAlarm(void);

static RETSIGTYPE TerminateHandler(int sig);

void X_Init(void);
void MainEventLoop(void);
void ReadFvwmPipe(void);
void LoopOnEvents(void);
void ShutMeDown(void);
int  ErrorHandler(Display *d, XErrorEvent *event);
void ProcessMessage(FvwmPacket* packet);
NetWindow *list_add(unsigned long *body);
NetWindow *list_configure(unsigned long *body, Bool *reset_wm_state);
void list_destroy(unsigned long *body);
int  list_focus(unsigned long *body);
int  list_new_page(unsigned long *body);
void list_restack(unsigned long *body, unsigned long length);
void list_iconify(unsigned long *body,unsigned long type);
void list_mini_icon(unsigned long *body);
void list_icon_file(unsigned long *body);
void list_window_name(unsigned long *body,unsigned long type);
void list_string(unsigned long *body);
void StrutPlacement(NetWindow *t);

/* ***************************************************************************
 * TerminateHandler - reentrant signal handler that ends the main event loop
 * *************************************************************************** */
static RETSIGTYPE TerminateHandler(int sig)
{
  /*
   * This function might not return - it could "long-jump"
   * right out, so we need to do everything we need to do
   * BEFORE we call it ...
   */
  fvwmSetTerminate(sig);
}

/*****************************************************************************
 *  Procedure: main - start of module
 * *************************************************************************** */
int main(int argc, char **argv)
{
  char *charset = NULL;
#ifdef HAVE_SIGACTION
  struct sigaction  sigact;
#endif

  /* Save the program name for error messages and config parsing */
  MyName = GetFileNameFromPath(argv[0]);

  if(argc != 6) {
    fprintf(stderr,"%s Version %s should only be executed by fvwm!\n",MyName,
	    VERSION);
    exit(1);
  }

  fd[0] = atoi(argv[1]);
  fd[1] = atoi(argv[2]);


#ifdef HAVE_SIGACTION
  sigemptyset(&sigact.sa_mask);
#ifdef SA_RESTART
    sigact.sa_flags = SA_RESTART;
#else
    sigact.sa_flags = 0;
#endif
    sigact.sa_handler = Alarm;
    sigaction(SIGALRM, &sigact, NULL);

#ifdef SA_INTERRUPT
  sigact.sa_flags = SA_INTERRUPT;
#else
  sigact.sa_flags = 0;
#endif
  sigemptyset(&sigact.sa_mask);
  sigaddset(&sigact.sa_mask, SIGTERM);
  sigaddset(&sigact.sa_mask, SIGPIPE);
  sigaddset(&sigact.sa_mask, SIGINT);
  sigact.sa_handler = TerminateHandler;
  sigaction(SIGPIPE, &sigact, NULL);
  sigaction(SIGTERM, &sigact, NULL);
  sigaction(SIGINT, &sigact, NULL);
#else
#ifdef USE_BSD_SIGNALS
  fvwmSetSignalMask( sigmask(SIGPIPE) |
                     sigmask(SIGINT)  |
                     sigmask(SIGTERM) );
#endif
  signal(SIGPIPE, TerminateHandler);
  signal(SIGTERM, TerminateHandler);
  signal(SIGINT, TerminateHandler);
  signal (SIGALRM, Alarm);
#ifdef HAVE_SIGINTERRUPT
  siginterrupt(SIGPIPE, True);
  siginterrupt(SIGTERM, True);
  siginterrupt(SIGINT, True);
#endif
#endif

  /* Initialize X connection + some variables init */
  X_Init();
  XSetErrorHandler(ErrorHandler);

  /* NET Initialisation (Scr.Net_w and ATOM init) */
  NET_Init();

  /* init the stacking window list and the Ping list*/
  InitList(&Stacking);
  InitList(&KdeSysTray);
  InitPingList(&PingList);

  /* get the charset */
  charset = UTF8_GetCharset();
  if (charset != NULL && strlen(charset) > 1)
     CopyString(&IconvCharset ,charset);

  /* Parse the config file */
  ParseConfig();

  /* initialisation of the imlib 1 or 1 */
#if HAVE_AN_IMLIB
  Iinit_Imlib();
#endif

  /* read the start files */
  ReadStartStopFiles("start",0);

  fd_width = GetFdWidth();

  /* Request a list of all windows,
   * wait for ConfigureWindow packets */
  SetMessageMask(fd,
#ifdef MINI_ICONS
		 M_MINI_ICON |
#endif
		 M_CONFIGURE_WINDOW |
		 M_ADD_WINDOW |
		 M_DESTROY_WINDOW |
		 M_WINDOW_NAME |
		 M_ICON_NAME |
		 M_RES_NAME |
		 M_RES_CLASS|
		 M_DEICONIFY |
		 M_ICONIFY |
		 M_END_WINDOWLIST |
		 M_NEW_DESK |
		 M_FOCUS_CHANGE |
		 M_RESTACK |
		 M_CONFIG_INFO |
		 M_SENDCONFIG |
		 M_NEW_PAGE |
		 M_STRING |
		 M_CONFIG_INFO |
		 M_ICON_FILE);

  SendFvwmPipe(fd, "Send_WindowList",0);

  /* Recieve all messages from Fvwm */
  atexit(ShutMeDown);

  /* set the desktop name and the number of desktop */
  LoadPersistentConfig();

  /* tell fvwm we're running */
  SendFinishedStartupNotification(fd);

  /* Lock on send only for iconify and deiconify (for NoIconAction) */
  SetSyncMask(fd, M_DEICONIFY | M_ICONIFY);
  SetNoGrabMask(fd, M_DEICONIFY | M_ICONIFY);

  MainEventLoop();
  return 0;
}

/* ***************************************************************************
 * X
 * *************************************************************************** */
void X_Init(void)
{
  char *display_name = NULL;

  if (!(dpy = XOpenDisplay(display_name)))
  {
    fprintf(stderr,"%s: can't open display %s", MyName,
	    XDisplayName(display_name));
    exit (1);
  }
  x_fd = XConnectionNumber(dpy);
  InitPictureCMap(dpy);

  Scr.Vx = 0;
  Scr.Vy = 0;
  Scr.CurrentDesk = 0;
  Scr.BottomLayer  = DEFAULT_BOTTOM_LAYER;
  Scr.DefaultLayer = DEFAULT_DEFAULT_LAYER;
  Scr.TopLayer     = DEFAULT_TOP_LAYER;

  Scr.screen = DefaultScreen(dpy);
  Scr.Root = RootWindow(dpy, Scr.screen);
  XSelectInput(dpy, Scr.Root, PropertyChangeMask | SubstructureNotifyMask);

  Scr.MyDisplayWidth = DisplayWidth(dpy, Scr.screen);
  Scr.MyDisplayHeight = DisplayHeight(dpy, Scr.screen);

  Scr.VxMax = 2*Scr.MyDisplayWidth;
  Scr.VyMax = 2*Scr.MyDisplayHeight;

  Scr.VxPages = Scr.VxMax / Scr.MyDisplayWidth;
  Scr.VyPages = Scr.VyMax / Scr.MyDisplayHeight;
  Scr.VWidth = Scr.VxMax + Scr.MyDisplayWidth;
  Scr.VHeight = Scr.VyMax + Scr.MyDisplayHeight;

  Scr.my_strut.left = 0;
  Scr.my_strut.right = 0;
  Scr.my_strut.top = 0;
  Scr.my_strut.bottom = 0;
}

/* ***************************************************************************
 * MainEventLoop -  Read and redraw until we die, blocking when can't read
 * *************************************************************************** */
void MainEventLoop(void)
{
  fd_set readset;

  while( !isTerminated ) {
    FD_ZERO(&readset);
    FD_SET(fd[1],&readset);
    FD_SET(x_fd,&readset);

    /* This code restyled after FvwmIconMan, which is simpler.
     * The biggest advantage over the original approach is
     * having one fewer select statements
     */
    XFlush(dpy);
    if (fvwmSelect(fd_width, &readset, NULL, NULL, NULL) > 0) {

      if (FD_ISSET(x_fd,&readset) || XPending(dpy)) LoopOnEvents();
      if (FD_ISSET(fd[1],&readset)) ReadFvwmPipe();

    }

  } /* while */
}


/* ***************************************************************************
 * ReadFvwmPipe - Read a single message from the pipe from Fvwm
 *   Originally Loop() from FvwmIdent:
 *     Copyright 1994, Robert Nation and Nobutaka Suzuki.
 * *************************************************************************** */
void ReadFvwmPipe(void)
{
    FvwmPacket* packet = ReadFvwmPacket(fd[1]);
    if ( packet == NULL )
	exit(0);
    else
	ProcessMessage( packet );
}

/* ***************************************************************************
 * LoopOnEvents - Process all the X events we get
 * *************************************************************************** */
void LoopOnEvents(void)
{
  XEvent Event;

  while(XPending(dpy))
  {
    XNextEvent(dpy,&Event);

    switch(Event.type)
    {
    case  PropertyNotify:
      NET_ProcessPropertyNotify(&Event);
      break;
    case ClientMessage:
      NET_ProcessClientMessage(&Event);
      break;
    default:
      break;
    }
  }

}

/******************************************************************************
  Alarm - Handle a SIGALRM - used to implement ping protocol
******************************************************************************/
RETSIGTYPE Alarm(int nonsense)
{
  PingItem *temp = (&PingList)->head;
  int t;
  NetWindow *tmp_nw;

  while(temp != NULL)
  {
    t = difftime(time(NULL), temp->p_time);
    if (t >= temp->p_time_out)
    {
      tmp_nw = GetNetWindow(temp->w);
      if (temp->p_action == PING_ACTION_INFO)
	fprintf(stderr,"%s: No answer for ping message with id: %i\n",
		MyName,temp->p_id);
      else if (temp->p_pid != 0) {
	if (temp->p_action == PING_ACTION_PING_KILL ||
	    temp->p_action == PING_ACTION_DESTROY)
	  kill(temp->p_pid, SIGKILL);
	else if (temp->p_action == PING_ACTION_PING_TERM)
	  kill(temp->p_pid, SIGTERM);
	else if (tmp_nw != NULL) {
	  if (temp->p_action == PING_ACTION_DELETE)
	    kill(temp->p_pid, SIGTERM);
	  else
	    kill(temp->p_pid, SIGKILL);
	}
      }
      else
      {
	/* not usefull a priori */
	SendText(fd,"Silent Destroy",temp->w);
      }
      DeletePingItem(&PingList,temp->w);
    }
    temp = temp->next;
  }

  if ((&PingList)->head == NULL)
    ClearAlarm();
  else
    SetAlarm();

#if !defined(HAVE_SIGACTION) && !defined(USE_BSD_SIGNALS)
  signal (SIGALRM, Alarm);
#endif
}

/***********************************************************************
 SetAlarm -- Schedule a timeout event
 ************************************************************************/
void SetAlarm()
{
  alarm(0);  /* remove a race-condition */
  alarm(1);
}

/***********************************************************************
 ClearAlarm -- Disable timeout events
 ************************************************************************/
void ClearAlarm(void)
{
  alarm(0);
}


/* ***************************************************************************
 * ShutMeDown - Do X client cleanup
 * *************************************************************************** */
void ShutMeDown(void)
{
  char command[1024];

  sprintf(command,"SendToModule fvwm2 FNHStop %s-stop %s-stop",
	  (FnhSysStartStopFile != NULL) ? FnhSysStartStopFile:"Null",
	  (FnhSysDesktopStartStopFile != NULL) ? FnhSysDesktopStartStopFile:"Null");
  SendText(fd,command,None);

  /* save the desktop names */
  SavePersistentConfig();
#if FREE_AT_EXIT
  {
    NetWindow *t,**prev;

    t = Start;
    prev = &Start;
    while(t != NULL) {
      if(prev != NULL)
	*prev = t->next;
      if(t->res_class != NULL)
	free(t->res_class);
      if(t->res_name != NULL)
	free(t->res_name);
      if(t->window_name != NULL)
	free(t->window_name);
      if(t->icon_name != NULL)
	free(t->icon_name);
      if (t->net_wm_name != NULL)
	free(t->net_wm_name);
      if (t->net_wm_icon_name != NULL)
	free(t->net_wm_icon_name);
      if(t->mini_icon_file != NULL)
	free(t->mini_icon_file);
      if(t->icon_file != NULL)
	free(t->icon_file);
#if FNH_IMLIB
      if(t->Imlib_mini != NULL) {
	if (t->mini_icon.picture != None)
	  Imlib_free_pixmap(ImlibId,t->mini_icon.picture);
	t->mini_icon.picture = None;
	t->mini_icon.mask = None;
	Imlib_destroy_image(ImlibId,t->Imlib_mini);
      }
#elif FNH_IMLIB2
      if(t->Imlib_mini != NULL) {
	imlib_context_set_image(t->Imlib_mini);
	imlib_free_image();
      }
#endif
      if (t->mini_icon.picture != None)
	XFreePixmap(dpy,t->mini_icon.picture);
      if (t->mini_icon.mask != None)
	XFreePixmap(dpy,t->mini_icon.mask);
      free(t);
      t = Start;
    }
  }
#endif
  XCloseDisplay(dpy);
}

/* ***************************************************************************
 * X Error Handler
 * *************************************************************************** */
int ErrorHandler(Display *d, XErrorEvent *event)
{
  /* some errors are OK=ish */
  if (event->error_code == BadPixmap)
    return 0;
  if (event->error_code == BadDrawable)
    return 0;
  /* seems this is needed by XGetWindowProperty (in NET_GetAtom) because
     the window may have disapear when we execute this command */
  if (event->error_code == BadWindow)
    return 0;
  
  PrintXErrorAndCoredump(d, event, MyName);
  return 0;
}

/* ***************************************************************************
 * time logger
 * *************************************************************************** */
#if FNH_DEBUG_TIME
void time_log(char *msg)
{
  clock_t time_val, time_taken;
  static clock_t start_time = 0;
  static clock_t prev_time = 0;
  struct tms not_used_tms;
  char buffer[200];                     /* oversized */
  time_t mytime;
  struct tm *t_ptr;
  static int counter = 0;

  time(&mytime);
  t_ptr = localtime(&mytime);
  if (start_time == 0) {
    prev_time = start_time =
      (unsigned int)times(&not_used_tms); /* get clock ticks */
  }
  time_val = (unsigned int)times(&not_used_tms); /* get clock ticks */
  time_taken = time_val - prev_time;
  prev_time = time_val;
  sprintf(buffer, "%.2d:%.2d:%.2d %6ld",
          t_ptr->tm_hour, t_ptr->tm_min, t_ptr->tm_sec, time_taken);
  fprintf(stderr,"%s[time]: %s at %s\n",MyName,msg,buffer);
}
#endif

/* ***************************************************************************
 * Procedure: Process message - examines packet types, and takes appropriate 
 * action
 * *************************************************************************** */
void ProcessMessage(FvwmPacket* packet)
{
  unsigned long type = packet->type;
  unsigned long length = packet->size;
  unsigned long* body = packet->body;
  NetWindow *netwin = NULL;
  int has_strut = 0;
  Bool reset_wm_state = False;
 
  switch(type)
  {
    case M_ADD_WINDOW:
    case M_CONFIGURE_WINDOW:
      if ((netwin = list_add(body)) != NULL) 
      {
	AddItem(&Stacking,body[0]);
	if (Started) {	  
	  NET_SetClientList();
	  NET_SetStackingList();
	  if (DO_HAS_STRUT(netwin)) {
	    NET_SetDesktopWorkArea();
	    if (UseDynamicStrutMaximize)
	      NET_UpdateMaximizedWindows();
	  }
	  if (UseStrutPlacement)
	    StrutPlacement(netwin);
	  NET_SetWMState(netwin);
	}
	if (UseExtendedName) {
	  NET_SetVisibleName(netwin,False);
	  NET_SetVisibleName(netwin,True);
	}
	XSelectInput(dpy, netwin->w, PropertyChangeMask);
      }
      else if ((netwin = list_configure(body,&reset_wm_state)) != NULL) {
	if (!IS_MAXIMIZED(netwin) && (netwin->s_max.horiz != 0 ||
				      netwin->s_max.vert != 0))
	  UpdateStrutMaximize(netwin->w,0,0,0,0,0,0);
	/* see if we have to reconfigure the window */
	NET_ReconfigureWindowType(netwin);
	if (reset_wm_state) {
	  NET_SetWMState(netwin);
	}
      }
      break;
    case M_DESTROY_WINDOW:
      list_destroy(body);
      break;
    case M_NEW_PAGE:
      if (list_new_page(body)) {
	Scr.CurrentDesk = body[2];
	NET_SetCurrentDesktop();
      }
      NET_SetDesktopViewPort();
      NET_SetDesktopGeometry();
      break;
    case M_NEW_DESK:
      if (Scr.CurrentDesk != body[0]) {
	Scr.CurrentDesk = body[0];
	NET_SetCurrentDesktop();
      }
      break;
    case M_FOCUS_CHANGE:
      if (list_focus(body)) 
	NET_SetActiveWindow();
      break;
    case M_ICONIFY:
    case M_DEICONIFY:
      /* fwvm will wait for an Unlock message before continuing
       * be careful when changing this construct, make sure unlock happens */
      list_iconify(body,type);
      SendUnlockNotification(fd);
      break;
    case M_RES_CLASS:
    case M_RES_NAME:
    case M_WINDOW_NAME:
    case M_ICON_NAME:
      list_window_name(body,type);
      break;
    case M_END_WINDOWLIST:
      if (!Started)
      {
	NET_SetAll();
	Started = 1;
#ifdef HAVE_AN_IMLIB
	SendMiniIcon(0);
#endif
      }
      break;
#ifdef MINI_ICONS
    case M_MINI_ICON:
      list_mini_icon(body);
      break;
#endif
    case M_ICON_FILE:
      list_icon_file(body);
      break;      
    case M_RESTACK:
      RestackList(&Stacking,body,length);
      NET_SetStackingList();
      NET_SetClientList();
      break;
    case M_CONFIG_INFO:
      {
	char *str;

	CopyString(&str, (char *)&body[3]);
	HandleConfig(str);
	if (str != NULL)
	  free(str);
      }
      break;
    case M_STRING:
      list_string(body);
      break;
    default:
      break;
  }
}

/* ***************************************************************************
 * Procedure: list_add - displays packet contents to stderr
 * *************************************************************************** */
NetWindow *list_add(unsigned long *body)
{
  NetWindow *t,**prev;
  struct ConfigWinPacket  *cfgpacket = (void *) body;
  int left,right,top,bottom;
  int do_skip_pager,is_stays_on_top,is_modal;

  t = Start;
  prev = &Start;

  while(t!= NULL)
  {
    if (t->w == cfgpacket->w)
    {
      /* it's already there, do nothing */
      return NULL;
    }
    prev = &(t->next);
    t = t->next;
  }

  *prev = (NetWindow *)safemalloc(sizeof(NetWindow));

  memset(*prev, 0, sizeof(NetWindow));
  (*prev)->w = cfgpacket->w;
  (*prev)->frame = cfgpacket->frame;
  (*prev)->icon_w = cfgpacket->icon_w;
  (*prev)->icon_pixmap_w = cfgpacket->icon_pixmap_w;
  (*prev)->x = cfgpacket->frame_x;
  (*prev)->y = cfgpacket->frame_y;
  (*prev)->width = cfgpacket->frame_width;
  (*prev)->height = cfgpacket->frame_height;

  (*prev)->frame_strut.left = (*prev)->frame_strut.right =
    cfgpacket->border_width;
  (*prev)->frame_strut.top = cfgpacket->border_width +
    ((HAS_BOTTOM_TITLE(cfgpacket)) ? 0 : cfgpacket->title_height);
  (*prev)->frame_strut.bottom = cfgpacket->border_width +
    ((HAS_BOTTOM_TITLE(cfgpacket)) ? cfgpacket->title_height : 0);
  NET_SetFrameStrut((*prev));

  (*prev)->desk = cfgpacket->desk;
  (*prev)->layer = cfgpacket->layer;
  (*prev)->default_layer = cfgpacket->layer;
  NET_GetWindowStrut(&left,&right,&top,&bottom,cfgpacket->w);
  (*prev)->strut_w.left   = (*prev)->dyn_strut_w.left   = left;
  (*prev)->strut_w.right  = (*prev)->dyn_strut_w.right  = right;
  (*prev)->strut_w.top    = (*prev)->dyn_strut_w.top    = top;
  (*prev)->strut_w.bottom = (*prev)->dyn_strut_w.bottom = bottom;
  (*prev)->s_max.horiz = (*prev)->s_max.h = 0;
  (*prev)->s_max.vert = (*prev)->s_max.v = 0;
  (*prev)->s_max.x = 0;
  (*prev)->s_max.y = 0;
  /* kicker set a negative strut when it is hidden (good idea)
   * but it will better than client can send a strut msg when  
   * it change it, moreover we need 2 struts: one "constant" that may
   * change (the user change the config) and one dynamic (the user hide
   * or shade the window) */ 
  if (left != 0 || top != 0 || right != 0 || bottom != 0)
    SET_DO_HAS_STRUT((*prev),True);
  else
    SET_DO_HAS_STRUT((*prev),False);
  memcpy(&((*prev)->flags), &(cfgpacket->flags), sizeof(cfgpacket->flags));

  /* HandleClientHints */
  (*prev)->net_window_type = NET_HandleWindowType((*prev));
  NET_HandleWMState((*prev),&do_skip_pager,&is_stays_on_top,&is_modal);
  SET_IS_NET_STAYS_ON_TOP((*prev),is_stays_on_top);
  SET_IS_NET_MODAL((*prev),is_modal);
  
  NET_HandleKdeSysTray(cfgpacket->w);

  if (PagerSkipListUseSkipList && DO_SKIP_WINDOW_LIST(cfgpacket))
    do_skip_pager = 1;
  SET_DO_SKIP_PAGER_WINDOW_LIST((*prev),do_skip_pager);

  SET_NET_MINI_ICON_OVERRIDE_MODE((*prev),MiniIconOverrideMode);
  SET_NET_ICON_OVERRIDE_MODE((*prev),IconOverrideMode);
  SET_NET_MINI_ICON_STATE((*prev),NET_GetWmIconState(cfgpacket->w,1));
  SET_NET_ICON_STATE((*prev),NET_GetWmIconState(cfgpacket->w,0));

  if ((*prev)->net_window_type == ATOM_NET_WM_WINDOW_TYPE_DESKTOP) {
    ReadStartStopFiles("start",1);
  }

  (*prev)->mini_icon.picture = None;
  (*prev)->mini_icon.mask = None;
  (*prev)->net_wm_visible_name_count = -1;
  (*prev)->net_wm_name = NULL;
  (*prev)->net_wm_icon_visible_name_count = -1;
  (*prev)->net_wm_icon_name = NULL;
#if HAVE_AN_IMLIB
  (*prev)->Imlib_mini = NULL;
  if (DO_OVERRIDE_NET_MINI_ICON((*prev))                  ||
      (DO_SIZE_OVERRIDE_NET_MINI_ICON((*prev))&&
       !DO_HAS_SIZE_NET_MINI_ICON((*prev)))                ||
      (DoSendMiniIcon == DO_NOT_SEND_NET_ICON)             ||
      DO_HAS_FVWM_NET_MINI_ICON((*prev))                   ||
      DO_HAS_NO_NET_MINI_ICON((*prev)))
  {
    /* nothing to do */
  }
  else
  {
    BuildMiniIcon((*prev), NULL, False);
  }
#endif /* HAVE_AN_IMLIB */

  return (*prev);
}

/* ***************************************************************************
 * Procedure: list_configure
 * *************************************************************************** */
NetWindow *list_configure(unsigned long *body, Bool *reset_wm_state)
{
  NetWindow *t;
  Window target_w;
  struct ConfigWinPacket  *cfgpacket = (void *) body;

  *reset_wm_state = False;
  target_w = cfgpacket->w;
  t = Start;
  while((t!= NULL)&&(t->w != target_w))
    t = t->next;

  if(t == NULL)
    return NULL;

  /* see if we have to set agian the _NET_WM_STATE */
  if (IS_STICKY(t) != IS_STICKY(cfgpacket)                      ||
      IS_MAXIMIZED(t) != IS_MAXIMIZED(cfgpacket)                ||
      IS_SHADED(t) != IS_SHADED(cfgpacket)                      ||
      DO_SKIP_WINDOW_LIST(t) != DO_SKIP_WINDOW_LIST(cfgpacket)  ||
      t->layer != cfgpacket->layer                              ||
      t->desk != cfgpacket->desk)
  {
    *reset_wm_state = True;
  }

  memcpy(&(t->flags), &(cfgpacket->flags), sizeof(cfgpacket->flags));
  t->desk = cfgpacket->desk;
  t->layer = cfgpacket->layer;
  t->frame = cfgpacket->frame;
  t->icon_w = cfgpacket->icon_w;
  t->icon_pixmap_w = cfgpacket->icon_pixmap_w;
  t->x = cfgpacket->frame_x;
  t->y = cfgpacket->frame_y;
  t->width = cfgpacket->frame_width;
  t->height = cfgpacket->frame_height;
  
  if (t->frame_strut.left != cfgpacket->border_width ||
      t->frame_strut.top != (cfgpacket->border_width +
      ((HAS_BOTTOM_TITLE(cfgpacket)) ? 0 : cfgpacket->title_height)) ||
      t->frame_strut.bottom != (cfgpacket->border_width +
      ((HAS_BOTTOM_TITLE(cfgpacket)) ? cfgpacket->title_height : 0))) 
  {
    t->frame_strut.left = t->frame_strut.right = cfgpacket->border_width;
    t->frame_strut.top = cfgpacket->border_width +
      ((HAS_BOTTOM_TITLE(cfgpacket)) ? 0 : cfgpacket->title_height);
    t->frame_strut.bottom = cfgpacket->border_width +
      ((HAS_BOTTOM_TITLE(cfgpacket)) ? cfgpacket->title_height : 0);
    NET_SetFrameStrut(t);
  }

  return t;
}

/* ***************************************************************************
 * Procedure: list_destroy - displays packet contents to stderr
 * *************************************************************************** */
void list_destroy(unsigned long *body)
{
  NetWindow *t,**prev;
  Window target_w;
  int has_strut = 0;

  target_w = body[0];
  t = Start;
  prev = &Start;
  while((t!= NULL)&&(t->w != target_w))
  {
    prev = &(t->next);
    t = t->next;
  }

  if (t == NULL)
    return;

  if (t->net_window_type == ATOM_NET_WM_WINDOW_TYPE_DESKTOP) {
    SendText(fd,"SendToModule fvwm2 DestroyDesktop",t->w);
    ReadStartStopFiles("stop",1);
  }
  if (DO_HAS_STRUT(t))
    has_strut = 1;
  if(prev != NULL)
    *prev = t->next;
  if(FocusWin == t)
    FocusWin = NULL;
  if(t->res_class != NULL)
    free(t->res_class);
  if(t->res_name != NULL)
    free(t->res_name);
  if(t->window_name != NULL)
    free(t->window_name);
  if (t->net_wm_name != NULL)
    free(t->net_wm_name);
  if(t->icon_name != NULL)
    free(t->icon_name);
  if (t->net_wm_icon_name != NULL)
    free(t->net_wm_icon_name);
  if(t->mini_icon_file != NULL)
    free(t->mini_icon_file);
  if(t->icon_file != NULL)
    free(t->icon_file);
#if FNH_IMLIB
  if(t->Imlib_mini != NULL) {
    if (t->mini_icon.picture != None)
      Imlib_free_pixmap(ImlibId,t->mini_icon.picture);
    t->mini_icon.picture = None;
    t->mini_icon.mask = None;
    Imlib_destroy_image(ImlibId,t->Imlib_mini);
  }
#elif FNH_IMLIB2
  if(t->Imlib_mini != NULL) {
    imlib_context_set_image(t->Imlib_mini);
    imlib_free_image();
  }
#endif
  if (t->mini_icon.picture != None)
    XFreePixmap(dpy,t->mini_icon.picture);
  if (t->mini_icon.mask != None)
    XFreePixmap(dpy,t->mini_icon.mask);
  free(t);

  DeleteItem(&Stacking,target_w);
  NET_SetClientList();
  NET_SetStackingList();
  if (has_strut) {
    NET_SetDesktopWorkArea();
    if (UseDynamicStrutMaximize)
      NET_UpdateMaximizedWindows();
  }

}

/* **************************************************************************
 *  Procedure: list_focus
 * *************************************************************************** */
int list_focus(unsigned long *body)
{
  NetWindow *t;
  Window target_w = body[0];

  t = Start;
  while((t!= NULL)&&(t->w != target_w))
  {
    t = t->next;
  }
  if(t != FocusWin)
  {
    FocusWin = t;
    return 1;
  }
  return 0;
}

/* ***************************************************************************
 * Procedure: list_new_page
 * *************************************************************************** */
int list_new_page(unsigned long *body)
{
  Scr.Vx = body[0];
  Scr.Vy = body[1];
  if((Scr.VxMax != body[3])||(Scr.VyMax != body[4]))
  {
    Scr.VxMax = body[3];
    Scr.VyMax = body[4];
    Scr.VxPages = Scr.VxMax / Scr.MyDisplayWidth;
    Scr.VyPages = Scr.VyMax / Scr.MyDisplayHeight;
    Scr.VWidth = Scr.VxMax + Scr.MyDisplayWidth;
    Scr.VHeight = Scr.VyMax + Scr.MyDisplayHeight;
  }
  if (Scr.CurrentDesk != body[2])
    return 1;
  else
    return 0;
}

/* ***************************************************************************
 * Procedure: list_iconify
 * *************************************************************************** */
void list_iconify(unsigned long *body,unsigned long type)
{
  NetWindow *t;
  Window w = body[0];

  t = Start;
  while((t!= NULL)&&(w != t->w))
  {
    t = t->next;
  }
  if (t == NULL)
    return;

  if ((type == M_DEICONIFY && IS_ICONIFIED(t))
      || (type == M_ICONIFY && !IS_ICONIFIED(t))) 
  {
    if (IS_ICON_SUPPRESSED(t)
	&& !DO_SKIP_WINDOW_LIST(t)
	&& NetAnimCommand != NULL
	&& NetAnimCommand[0] != 0)
    {
      int x,y,width,height;
      
      NET_GetWmIconGeometry(&x,&y,&width,&height,body[0]);
      if (0 <= x <= x+width && x+width <= Scr.MyDisplayWidth  && 
	  0 <= y <= y+height &&  y+height <= Scr.MyDisplayHeight)
      {
	char buff[MAX_MODULE_INPUT_TEXT_LEN];
	  
	if (IS_ICONIFIED(t)) {
	  sprintf(buff, "%s %d %d %d %d %d %d %d %d", NetAnimCommand,
		  x, y, width, height,
		  (int)body[7], (int)body[8], (int)body[9], (int)body[10]);
	} else {
	  sprintf(buff, "%s %d %d %d %d %d %d %d %d", NetAnimCommand,
		  (int)body[7], (int)body[8], (int)body[9], (int)body[10],
		  x, y, width, height);
	}
	SendText(fd, buff, 0);
      }
    }
  }
  if (type == M_DEICONIFY)
    SET_ICONIFIED(t, False);
  else
    SET_ICONIFIED(t, True);
}

/* ***************************************************************************
 * Procedure: list_mini_icon
 * *************************************************************************** */
void list_mini_icon(unsigned long *body)
{
  NetWindow	*t;
  Window target_w=body[0];
  char *path;

  /* this mini icon is one I've set: may be a litle problem at restart */
  if (!strcmp((char *)(&body[8]),"FvwmNetHints")) {
    return;
  }

  t = Start;
  while (t && (t->w != target_w))
    t = t->next;
  if (t == NULL)
    return;

  if (t->mini_icon_file != NULL) {
    free(t->mini_icon_file);
    t->mini_icon_file = NULL;
  }

  CopyString(&t->mini_icon_file,(char *)(&body[8]));

  if (!DO_OVERRIDE_NET_MINI_ICON(t)                                       &&
      !(DO_SIZE_OVERRIDE_NET_MINI_ICON(t)&&!DO_HAS_SIZE_NET_MINI_ICON(t)) &&
      !DO_HAS_FVWM_NET_MINI_ICON(t)                                        &&
      !DO_HAS_NO_NET_MINI_ICON(t))
  {
#if HAVE_AN_IMLIB
    if (t->Imlib_mini != NULL)
      SendMiniIcon(t->w);
#endif
    return;
  }

  if (DoSendMiniIcon == DO_NOT_SEND_NET_ICON && DO_NOT_ADD_NET_MINI_ICON(t))
    return;

  path = findImageFile(t->mini_icon_file, ImagePath, R_OK );
  
  if (path == NULL)
    return;
  
#if XPM_FOR_FNH
  if (!DO_NOT_ADD_NET_MINI_ICON(t) && NET_SetWmIconFromfile(target_w, path, 1))
  {
      SET_NET_MINI_ICON_STATE(t,FVWM_NET_ICON);
  }
#endif

#if HAVE_AN_IMLIB
  if (DoSendMiniIcon == DO_SEND_NET_ICON_ALL) {
    if (BuildMiniIcon(t, path, False))
      SendMiniIcon(t->w);
  }
#endif
}

/* ***************************************************************************
 * Procedure: list_icon_file
 * *************************************************************************** */
void list_icon_file(unsigned long *body)
{
  NetWindow	*t;
  Window target_w=body[0];
  char *path;
  
  t = Start;
  while (t && (t->w != target_w))
    t = t->next;
  if (t == NULL)
    return;

  if (t->icon_file != NULL) {
    free(t->icon_file);
    t->icon_file = NULL;
  }
  
  CopyString(&t->icon_file,(char *)(&body[3]));

  if (!DO_OVERRIDE_NET_ICON(t)                                  &&
      !(DO_SIZE_OVERRIDE_NET_ICON(t)&&!DO_HAS_SIZE_NET_ICON(t)) &&
      !DO_HAS_FVWM_NET_ICON(t)                                   &&
      !DO_HAS_NO_NET_ICON(t))
  {
    return;
  }

  if (DO_NOT_ADD_NET_ICON(t))
    return;

  path = findImageFile(t->icon_file, ImagePath, R_OK );

  if (path == NULL)
    return;

#if XPM_FOR_FNH
  if (NET_SetWmIconFromfile(target_w, path, 0)) {
    SET_NET_ICON_STATE(t,FVWM_NET_ICON);
  }
#endif
}

/* ***************************************************************************
 *
 * *************************************************************************** */
void list_window_name(unsigned long *body,unsigned long type)
{
  NetWindow *t;
  Window target_w;

  target_w = body[0];
  t = Start;
  while((t!= NULL)&&(t->w != target_w))
  {
    t = t->next;
  }
  if(t == NULL)
    return;

  switch (type) {
  case M_RES_CLASS:
    if(t->res_class != NULL)
      free(t->res_class);
    CopyString(&t->res_class,(char *)(&body[3]));
    break;
  case M_RES_NAME:
    if(t->res_name != NULL)
      free(t->res_name);
    CopyString(&t->res_name,(char *)(&body[3]));
    break;
  case M_WINDOW_NAME:
    if(t->window_name != NULL)
      free(t->window_name);
    CopyString(&t->window_name,(char *)(&body[3]));
#if HAVE_ICONV
    if (UseExtendedName &&
	IconvCharset != NULL &&
	!DO_HAS_NET_WM_NAME(t) &&
	t->window_name != NULL)
    { 
      if (t->net_wm_name != NULL) {
	free(t->net_wm_name);
	t->net_wm_name = NULL;
      }
      t->net_wm_name = 
	(unsigned char *)UTF8_charset2utf8(IconvCharset,t->window_name);
      if (t->net_wm_name != NULL)
	NET_SetVisibleName(t,False);
    }
#endif
    break;
  case M_ICON_NAME:
    if(t->icon_name != NULL)
      free(t->icon_name);
    CopyString(&t->icon_name,(char *)(&body[3]));
#if HAVE_ICONV
    if (UseExtendedName &&
	IconvCharset != NULL &&
	!DO_HAS_NET_WM_ICON_NAME(t) &&
	t->icon_name != NULL)
    {
      if (t->net_wm_icon_name != NULL) {
	free(t->net_wm_icon_name);
	t->net_wm_icon_name = NULL;
      }
      t->net_wm_icon_name = 
	(unsigned char *)UTF8_charset2utf8(IconvCharset,t->icon_name);
      if (t->net_wm_icon_name != NULL)
	NET_SetVisibleName(t,True);
    }
#endif
    break;
  }

  /* ksmserver workaround */
  if (KsmServerWorkAround &&
      t->window_name != NULL && t->res_name != NULL &&
      t->icon_name != NULL && t->res_class != NULL &&
      strcmp(t->window_name,"ksmserver") == 0 &&
      strcmp(t->res_class,"ksmserver") == 0 &&
      strcmp(t->icon_name,"ksmserver") == 0 &&
      strcmp(t->res_name,"unnamed") == 0)
  {
    char cmd[256];
    int layer = 0;
    
    if (IS_TRANSIENT(t) &&
	  t->width < Scr.MyDisplayWidth - 100 &&
	t->height < Scr.MyDisplayHeight - 100)
      layer = StaysOnTopLayer + 2;
    else if (t->width == Scr.MyDisplayWidth && 
	     t->height == Scr.MyDisplayHeight)
      layer = StaysOnTopLayer + 1;
    
    if (layer) {
      sprintf(cmd,"Silent Stick On, Silent Layer 0 %i",layer);
      SendFvwmPipe(fd,cmd,t->w);
      }
  }

}

/* ***************************************************************************
 * list_string
 * *************************************************************************** */
void list_string(unsigned long *body)
{
  char *str;
  char *token;
  char *next;
  Window app_win = None;
  NetWindow *netwin = NULL;

  CopyString(&str, (char *)&body[3]);
  token = PeekToken(str, &next);

  /**** MAXIMIZE ****/
  if (StrEquals(token, "Maximize"))
  {
    int horiz = DefaultHorizMaximize;
    int vert  = DefaultVertMaximize;
    int val;
    int toggle;
    int dummy = 0;
    
    toggle = ParseToggleArgument(next, &next, -1, 0);
    if (toggle == 1)
      toggle = _NET_WM_STATE_ADD;
    else if (toggle == 0)
      toggle = _NET_WM_STATE_REMOVE;
    else if (toggle == -1)
      toggle = _NET_WM_STATE_TOGGLE;
    
    token = PeekToken(next, &next);
    if (GetOnePercentArgument(token, &val, &dummy) && dummy != 100) {
      if (val < 0)
	horiz = 100 + val;
      else
	horiz = val;
    }
    token = PeekToken(next, NULL);
    if (GetOnePercentArgument(token, &val, &dummy) && dummy !=100) {
      if (val < 0)
	vert = 100 + val;
      else
	vert = val;
    }
    if (body[0] != 0) {
      app_win = body[0];
    } else {
      fvwmlib_get_target_window(dpy, Scr.screen, MyName, &app_win, True);
    }
    netwin = GetNetWindow(app_win);
    if (netwin != NULL && !IS_ICONIFIED(netwin) &&
	netwin->net_window_type != ATOM_NET_WM_WINDOW_TYPE_DESKTOP)
      NET_StrutMaximize(horiz, vert, toggle, netwin);
  }
  /***** PING *****/
  else if (StrEquals(token,"Ping"))
  {
    int action = PING_ACTION_INFO;
    int time_out = 5;
    
    token = PeekToken(next, &next);
    if (StrEquals(token,"Info"))
      action = PING_ACTION_INFO;
    else if (StrEquals(token,"Delete"))
      action = PING_ACTION_DELETE;
    else if (StrEquals(token,"Close"))
      action = PING_ACTION_CLOSE;
    else if (StrEquals(token,"Destroy"))
      action = PING_ACTION_DESTROY;
    else if (StrEquals(token,"PingKill"))
      action = PING_ACTION_PING_KILL;
    else if (StrEquals(token,"PingTerm"))
      action = PING_ACTION_PING_TERM;
    
    if (body[0] != 0) {
      app_win = body[0];
    } else {
      fvwmlib_get_target_window(dpy, Scr.screen, MyName, &app_win, True);
    }
    netwin = GetNetWindow(app_win);
    if (netwin != NULL &&
	netwin->net_window_type != ATOM_NET_WM_WINDOW_TYPE_DESKTOP)
      NET_PingWindow(netwin,action,time_out);
  }
  /**** PCONFIG ****/
  else if (StrEquals(token, "ReloadPConfig"))
    LoadPersistentConfig();
  else if (StrEquals(token, "SavePConfig"))
    SavePersistentConfig();
  /**** START Files ****/
  else if (StrEquals(token, "ReadStartFile")) {
    ReadStartStopFiles("start",0);
  }
  else if (StrEquals(token, "ReadStartDesktopFile")) {
    NetWindow *t;

    t = Start;
    while(t != NULL && t->net_window_type != ATOM_NET_WM_WINDOW_TYPE_DESKTOP)
    {
      t = t->next;
    }
    if(t != NULL) {
      ReadStartStopFiles("start",1);
    }
  }
  else if (StrEquals(token, "Quit")) {
    exit(0);
  }
  else
    fprintf(stderr,"%s: Unknown command: %s\n",MyName,token);
  
  if (str != NULL)
    free(str);
}

/* ***************************************************************************
 *
 * *************************************************************************** */
NetWindow *GetNetWindow(Window w)
{
  NetWindow *t;

  if (w == None)
    return NULL;

  t = Start;
  while(t!= NULL &&
	w != t->w && 
	w != t->frame &&
	w != t->icon_w &&
	w != t->icon_pixmap_w)
  {
    t = t->next;
  }
  if(t != NULL)
    return t;

  return NULL;
}

/* ***************************************************************************
 * 
 * *************************************************************************** */
void UpdateSkipPager(Window w, Bool state)
{
  NetWindow *t;

  t = Start;
  while((t!= NULL)&&(w != t->w))
  {
    t = t->next;
  }
  if(t != NULL)
    SET_DO_SKIP_PAGER_WINDOW_LIST(t,state);
}

/* ***************************************************************************
 * 
 * *************************************************************************** */
void UpdateNetModal(Window w, Bool state)
{
  NetWindow *t;

  t = Start;
  while((t!= NULL)&&(w != t->w))
  {
    t = t->next;
  }
  if(t != NULL)
    SET_IS_NET_MODAL(t,state);
}
/* ***************************************************************************
 * 
 * *************************************************************************** */
void UpdateStrutMaximize(Window w, int horiz, int vert, int h, int v,
			 int x, int y)
{
  NetWindow *t;

  t = Start;
  while((t!= NULL)&&(w != t->w))
  {
    t = t->next;
  }
  if(t != NULL) {
    t->s_max.horiz = horiz;
    t->s_max.vert = vert;
    t->s_max.h = h;
    t->s_max.v = v;
    t->s_max.x = x;
    t->s_max.y = y;
  }
}

#ifdef HAVE_AN_IMLIB
/* ***************************************************************************
 *
 * ***************************************************************************/
#define DO_USE_COLOR_LIMIT 0
void SendMiniIcon(Window w)
{
  NetWindow *t;
  char command[256];

  if (!DoSendMiniIcon)
    return;

  t = Start;
  while(t != NULL) {
    if (t->w == w || w == 0) {
      if (t->Imlib_mini != NULL && t->mini_icon.picture != None) {
	sprintf(command,"SendToModule fvwm2 MiniIcon %lx %lx %i FvwmNetHints",
		t->mini_icon.picture, t->mini_icon.mask, DO_USE_COLOR_LIMIT);
	SendText(fd,command,t->w);
      }
    }
    t = t->next;
  }
}

/* ***************************************************************************
 * Build a mini icon from a icon net hint 
 * ***************************************************************************/
Bool BuildMiniIcon(NetWindow *t, char *path, Bool rebuild)
{
  int im_w,im_h;

  if (t == NULL)
    return 0;

  if (rebuild && t->Imlib_mini == NULL)
    return 0;

  if (t->Imlib_mini != NULL) {
#if FNH_IMLIB
    if (t->mini_icon.picture != None)
      Imlib_free_pixmap(ImlibId,t->mini_icon.picture);
    t->mini_icon.picture = None;
    t->mini_icon.mask = None;
    if (!rebuild)
      Imlib_destroy_image(ImlibId,t->Imlib_mini);
#else /* FNH_IMLIB2 */
    imlib_context_set_image(t->Imlib_mini);
    if (!rebuild) {
      imlib_free_image();
    }
#endif
  }

  if (t->mini_icon.picture != None)
    XFreePixmap(dpy,t->mini_icon.picture);
  if (t->mini_icon.mask != None)
    XFreePixmap(dpy,t->mini_icon.mask);

  if (!rebuild) {
    if (path != NULL)
    {
#if FNH_IMLIB
      if (t->Imlib_mini == NULL)
	t->Imlib_mini = Imlib_load_image(ImlibId,path);
#else /* FNH_IMLIB2 */
      t->Imlib_mini = imlib_load_image(path);
      imlib_context_set_image(t->Imlib_mini);
      imlib_image_set_has_alpha(1);
#endif
    }
    else
      t->Imlib_mini = NET_GetNetWmIcons(t->w);
  }

  if (t->Imlib_mini == NULL)
    return 0;

#if FNH_IMLIB
  im_w = (MiniIconSize[0]>0) ? MiniIconSize[0] : t->Imlib_mini->rgb_width;
  im_h = (MiniIconSize[1]>0) ? MiniIconSize[1] : t->Imlib_mini->rgb_height;
  Imlib_render(ImlibId,t->Imlib_mini,im_w,im_h);
  t->mini_icon.picture = Imlib_move_image(ImlibId,t->Imlib_mini);
  t->mini_icon.mask = Imlib_move_mask(ImlibId,t->Imlib_mini);
  t->mini_icon.width = im_w;
  t->mini_icon.height = im_h;
  t->mini_icon.depth = ImlibId->x.depth;
#else /* FNH_IMLIB2 */
  im_w = (MiniIconSize[0]>0) ? MiniIconSize[0] : imlib_image_get_width();
  im_h = (MiniIconSize[1]>0) ? MiniIconSize[1] : imlib_image_get_height();
  t->mini_icon.width = im_w;
  t->mini_icon.height = im_h;
  t->mini_icon.depth = Pdepth;
  t->mini_icon.picture = XCreatePixmap(dpy,Scr.Net_w,im_w,im_h,Pdepth);
  t->mini_icon.mask = XCreatePixmap(dpy,Scr.Net_w,im_w,im_h,1);
  imlib_context_set_drawable(t->mini_icon.picture);
  imlib_context_set_mask(t->mini_icon.mask);
  imlib_render_image_on_drawable_at_size(0,0,im_w,im_h);
#endif
  return 1;
}

#endif /* HAVE_AN_IMLIB */

/* ***************************************************************************
 *
 * ***************************************************************************/
void StrutPlacement(NetWindow *t)
{
  int left,right,top,bottom;
  int new_x = 0;
  int new_y = 0;
  char cmd[256];

  if (!IS_PLACED_BY_FVWM(t))
    return;
  NET_GetStrut(&left, &right, &top, &bottom, t->desk, 0);

  if (left > t->x && (left + t->width <= Scr.MyDisplayWidth - right))
    new_x = (left - t->x);
  else if (t->x + t->width > Scr.MyDisplayWidth - right &&
	   left + t->width <= Scr.MyDisplayWidth - right)
    new_x = Scr.MyDisplayWidth - right - t->width - t->x;

  if (top > t->y && top + t->height <= Scr.MyDisplayHeight - bottom)
    new_y = (top - t->y);
  else if (t->y + t->height > Scr.MyDisplayHeight - bottom &&
	   top + t->height <= Scr.MyDisplayHeight - bottom)
    new_y = - t->y - t->height + Scr.MyDisplayHeight - bottom;

  if (new_x != 0 || new_y != 0)
  {
    sprintf(cmd, "Silent Move w%ip w%ip", new_x, new_y);
    SendText(fd,cmd,t->w);
  }
}
