/*
 * Copyright 2001, Olivier Chapuis. No guarantees or warantees or anything
 * are provided or implied in any way whatsoever. Use this program at your
 * own risk. Permission to use this program for any purpose is given,
 * as long as the copyright is kept intact. */

/* This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#ifdef SHAPE
#include <X11/extensions/shape.h>
#endif
#include <X11/Xatom.h>
#include <X11/Xproto.h>

#include "my_imlib.h"
#include "Mallocs.h"
#include "libs/Module.h"

#include "FvwmNetHints.h"
#include "net.h"
#include "net_debug.h"

extern Display *dpy;
extern ScreenInfo Scr;

extern int fd[2];
extern char *MyName;
extern Atom ATOM_NET_WM_ICON;
extern Atom ATOM_FVWM_NET_SET_MINI_ICON;
extern Atom ATOM_FVWM_NET_SET_ICON;
extern unsigned MiniIconMaxSize;
#if FNH_IMLIB
extern ImlibData *ImlibId;
#endif

/* ***************************************************************************
 *
 * ***************************************************************************/
unsigned NET_GetWmIconState(Window w,int is_mini_icon)
{
  CARD32 *list;
  CARD32 *d;
  unsigned int list_size;
  int r = NO_NET_ICON;
  unsigned int i = 0;
  unsigned int s;

  list = NET_AtomGet(w, ATOM_NET_WM_ICON, XA_CARDINAL, &list_size);
  if (list == NULL)
    return r;
 
  r = NET_ICON;
  s = (list_size / (sizeof(CARD32)));

  while(i < s)
  {
    if (i+1 < s)
    {
      if (i + 1 + list[i]*list[i+1] < s)
      {
	if (!is_mini_icon && (list[i] > MiniIconMaxSize || 
			     list[i+1] >  MiniIconMaxSize))
	{
	  r = SIZE_NET_ICON;
	  i = s;
	}
	if (is_mini_icon && list[i] <= MiniIconMaxSize && 
	    list[i+1] <= MiniIconMaxSize)
	{
	  r = SIZE_NET_ICON;
	  i = s;
	}
      }
      if (i != s && list[i]*list[i+1] > 0)
	i = i + 2 + list[i]*list[i+1];
      else
	i = s;
    }
    else
      i = s;
  } /* while */
  free(list);
  if (r == SIZE_NET_ICON) {
    if ((is_mini_icon && 
	 (d=NET_AtomGet(w,ATOM_FVWM_NET_SET_MINI_ICON,XA_CARDINAL,&list_size)) 
	 != NULL) || 
	(!is_mini_icon &&
	 (d=NET_AtomGet(w,ATOM_FVWM_NET_SET_ICON, XA_CARDINAL, &list_size)) 
	 != NULL)) 
      r = FVWM_NET_ICON;
  } else {
    /* r = NET_ICON */
    if ((is_mini_icon && 
	 (d=NET_AtomGet(w, ATOM_FVWM_NET_SET_ICON, XA_CARDINAL, &list_size))
	!= NULL) ||
	(!is_mini_icon && 
	 (d=NET_AtomGet(w,ATOM_FVWM_NET_SET_MINI_ICON,XA_CARDINAL, &list_size)) 
	!= NULL)) {
      r= NO_NET_ICON;
    }
  }
  if (d != NULL)
    free(d);
  return r;
}


/* ***************************************************************************
 * Replace/Add the NET icon of w. If is_mini_icon is True and if w has a
 * a NET icon with width or height >  MiniIconMaxSize, then this icon
 * is conserved and the one in file. If is_mini_icon is False the first icon
 * of size <= MiniIconMaxSizexMiniIconMaxSize is conserved.
 * In any case the icon in file is added (if possible) to the above icon
 * if any. So, we support only 2 icons. We return True if we encountre no
 * problems.
 * ***************************************************************************/
#if XPM_FOR_FNH
int NET_SetWmIconFromfile(Window w, char *file, int is_mini_icon)
{
  CARD32 *orig_list;
  unsigned int orig_list_size;
  int keep_start = 0;
  int keep_length = 0;
  XpmImage xpm = {0};
  unsigned int i,j;
  XColor rgb;
  unsigned int my_rgb[3];

  if (XpmReadFileToXpmImage(file, &xpm, NULL) != XpmSuccess)
  {
    XpmFreeXpmImage(&xpm);
    return 0;
  }

  /* see if we have already set an icon or mini icon */ 
  orig_list = NET_AtomGet(w, ATOM_NET_WM_ICON, XA_CARDINAL, &orig_list_size);

  if (orig_list != NULL) {
    int s = (orig_list_size / (sizeof(CARD32)));

    i = 0;
    while(i < s)
    {
      if (i+1 < s)
      {
	if (i + 1 + orig_list[i]*orig_list[i+1] < s)
	{
	 if (is_mini_icon && (orig_list[i] > MiniIconMaxSize || 
			      orig_list[i+1] >  MiniIconMaxSize))
	 {
	   keep_start = i;
	   keep_length =  2 + orig_list[i]*orig_list[i+1];
	   i = s;
	 }
	 if (!is_mini_icon && orig_list[i] <= MiniIconMaxSize && 
	     orig_list[i+1] <= MiniIconMaxSize)
	 {
	   keep_start = i;
	   keep_length =  2 + orig_list[i]*orig_list[i+1];
	   i = s;
	 }
	}
	if (i != s && orig_list[i]*orig_list[i+1] > 0)
	  i = i + 2 + orig_list[i]*orig_list[i+1];
	else
	  i = s;
      }
      else
	i = s;
    }
  }
  
  /* now build the NET icon */
  {
    CARD32 list[xpm.height*xpm.width+2+keep_length];
    Atom a;
    CARD32 val = MiniIconMaxSize;
    int k,l;
    unsigned int argbTable[xpm.ncolors];

    /* build the "net" color table */
    for(i = 0; i < xpm.ncolors; i++) {
      if (xpm.colorTable[i].c_color == NULL || 
	  !strcasecmp(xpm.colorTable[i].c_color,"none"))
      {
	argbTable[i] = 0;
      }
      else if (XParseColor(dpy, Pcmap, xpm.colorTable[i].c_color,&rgb) != 0)
      {
	my_rgb[0] = rgb.blue;
	my_rgb[1] = rgb.green;
	my_rgb[2] = rgb.red;
	/* asuume 16 bits or 8 bits */
	argbTable[i] = 0;
	for (l=0;l<3;l++) {
	  if (my_rgb[l] > 255) {
	    my_rgb[l] = my_rgb[l]/256;
	  }
	  argbTable[i] += my_rgb[l] << 8*l;
	}
	argbTable[i] += (unsigned)(255 << 24);
      }
      else
      {
	argbTable[i] = 0;
      }
    }

    for(i = 0; i < keep_length; i++)
      list[i] = orig_list[keep_start+i];

    list[keep_length] = xpm.width;
    list[1+keep_length] = xpm.height;
   
    for (i = 0; i < xpm.height; i++) {
      for (j = 0; j < xpm.width; j++) {
	k = j+i*xpm.width;
	list[k+2+keep_length] = argbTable[xpm.data[k]];
      }
    }

    XChangeProperty(dpy,w,ATOM_NET_WM_ICON, XA_CARDINAL,
		    32, PropModeReplace, (unsigned char *)&list,
		    xpm.height*xpm.width+2+keep_length);
    if (is_mini_icon)
      a = ATOM_FVWM_NET_SET_MINI_ICON;
    else
      a = ATOM_FVWM_NET_SET_ICON;
    XChangeProperty(dpy,w,a, XA_CARDINAL, 32, PropModeReplace,
		    (unsigned char *)&val,1);
  }

  if (orig_list != NULL)
    free(orig_list);

  XpmFreeXpmImage(&xpm);
  return 1;
}
#endif /* XPM_FOR_FNH */

#if HAVE_AN_IMLIB
/* ***************************************************************************
 * Initialisation of the Imlib
 * ***************************************************************************/
void Iinit_Imlib(void)
{
#if FNH_IMLIB
  ImlibId = Imlib_init(dpy);
#else /* FNH_IMLIB2 */
  /* image cache (in bytes) */
  imlib_set_cache_size(1024 * 512);
   /* set the maximum number of colors to allocate for 8bpp */
  imlib_set_color_usage(256);   
  /* dither for depths < 24bpp */
  imlib_context_set_dither(1);
  /* mask */
  imlib_context_set_dither_mask(0);
   /* */
  imlib_context_set_blend(0);
   /* set the display , visual, colormap and drawable we are using */
  imlib_context_set_display(dpy);
  imlib_context_set_visual(Pvisual);        
  imlib_context_set_colormap(Pcmap);
#endif
}

/* ***************************************************************************
 * Create an imlib image from a NET icon
 * ***************************************************************************/
myImlibImage *NET_GetNetWmIcons(Window w)
{
  CARD32 *list;
  unsigned int size;
  unsigned int i = 0;
  unsigned int start_best = 0;
  int best_w = 0;
  int best_h = 0;
  myImlibImage *im = NULL;
  
  list = NET_AtomGet(w, ATOM_NET_WM_ICON, XA_CARDINAL, &size);
  if (list == NULL)
    return NULL;

  size = size / (sizeof(CARD32));

  while (i < size)
  {
    if (i+1 < size)
    {
      if (i + 1 + list[i]*list[i+1] < size)
      {
	if (best_w == 0 && best_h == 0) {
	  start_best = i+2;
	  best_w = list[i];
	  best_h = list[i+1];
	}
	else if (list[i]*list[i+1] < best_w*best_h)
	{
	  start_best = i+2;
	  best_w = list[i];
	  best_h = list[i+1];
	}
      }
      if (list[i]*list[i+1] > 0)
	i = i + 2 + list[i]*list[i+1];
      else
	i = size;
    }
    else
      i = size;
  } /* while */

  if (start_best == 0 || best_w+best_h == 0) {
    free(list);
    return NULL;
  }
  
  {
    unsigned char *char_list = (unsigned char *)list;
    int j = 0;
#if FNH_IMLIB
    unsigned char data[3*best_w*best_h];
    ImlibColor color;
#else /* FNH_IMLIB2 */
     unsigned char data[4*best_w*best_h];
#endif

#if FNH_IMLIB 
    color.r = 255;
    color.g = 1;
    color.b = 255;
#endif

    for(i=0; i<best_w*best_h*4;i = i + 4) {
#if FNH_IMLIB
      if ((unsigned int)char_list[i + 3 + 4*start_best] >= 130)
      {
	data[j] = char_list[i + 2 + 4*start_best];
	data[j+1] = char_list[i + 1 + 4*start_best];
	data[j+2] = char_list[i + 0 + 4*start_best];
	if ((unsigned int)data[j] == 255 &&
	    (unsigned int)data[j+1] == 1 &&
	    (unsigned int)data[j+2] == 255)
	  (unsigned int)data[j+1] = 0;
      }
      else
      {
	data[j] = 255;
	data[j+1] = 1;
	data[j+2] = 255;
      }
      j=j+3;
#else /* FNH_IMLIB2 */
      data[j] = char_list[i + 0 + 4*start_best];
      data[j+1] = char_list[i + 1 + 4*start_best];
      data[j+2] = char_list[i + 2 + 4*start_best];
      if ((unsigned int)char_list[i + 3 + 4*start_best] >= 130)
      {
	      data[j+3] = 255;
      }
      else
      {
	      data[j+3] = 0;
      }
      j=j+4;
#endif
    }

#if FNH_IMLIB
    im = Imlib_create_image_from_data(ImlibId, (unsigned char *)data,NULL,
				      best_w, best_h);
    Imlib_set_image_shape(ImlibId, im, &color);
#else /* FNH_IMLIB2 */
    im =
      imlib_create_image_using_copied_data(best_w,best_h,(DATA32 *)data);
    imlib_context_set_image(im);
    imlib_image_set_has_alpha(1);
#endif
  }

  free(list);

  if (im != NULL) {
    return im;
  }

  return NULL;
}

#endif /* HAVE_AN_IMLIB */

