/*
 * Copyright 2001, Olivier Chapuis. No guarantees or warantees or anything
 * are provided or implied in any way whatsoever. Use this program at your
 * own risk. Permission to use this program for any purpose is given,
 * as long as the copyright is kept intact. */

/* This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "config.h"

#include <X11/Intrinsic.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xproto.h>
#include <X11/Xatom.h>


#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "FvwmNetHints.h"
#include "net.h"
#include "windowlist.h"
#include "Mallocs.h"
#include "net_debug.h"

#include "libs/Module.h"
#include "fvwm/fvwm.h"
extern char *MyName;
extern int fd[2];

extern ScreenInfo Scr;
extern Display *dpy;
extern NetWindow *Start;
extern NetWindow *FocusWin;
extern WindowList Stacking;
extern WindowList KdeSysTray;
extern PingWindowList PingList;
extern long NumberOfDesktops;
extern int Started;

#define XIA(a) XInternAtom(dpy,a,False);

/* Root Window properties (+ Related Message) */
Atom ATOM_NET_SUPPORTED            = None;
Atom ATOM_NET_CLIENT_LIST          = None;
Atom ATOM_NET_CLIENT_LIST_STACKING = None;
Atom ATOM_NET_NUMBER_OF_DESKTOPS   = None;
Atom ATOM_NET_DESKTOP_GEOMETRY     = None;
Atom ATOM_NET_DESKTOP_VIEWPORT     = None;
Atom ATOM_NET_CURRENT_DESKTOP      = None;
Atom ATOM_NET_DESKTOP_NAMES        = None;
Atom ATOM_NET_ACTIVE_WINDOW        = None;
Atom ATOM_NET_WORKAREA             = None;
Atom ATOM_NET_SUPPORTING_WM_CHECK  = None;
Atom ATOM_NET_VIRTUAL_ROOTS        = None;

/* Other Root window message */
Atom ATOM_NET_CLOSE_WINDOW  = None;
Atom ATOM_NET_WM_MOVERESIZE = None;

/* Applications windows properties */
Atom ATOM_NET_WM_NAME                  = None;
Atom ATOM_NET_WM_VISIBLE_NAME          = None;
Atom ATOM_NET_WM_ICON_NAME             = None;
Atom ATOM_NET_WM_ICON_VISIBLE_NAME     = None;
Atom ATOM_NET_WM_DESKTOP               = None;

Atom ATOM_NET_WM_WINDOW_TYPE           = None;
Atom ATOM_NET_WM_WINDOW_TYPE_DESKTOP   = None;
Atom ATOM_NET_WM_WINDOW_TYPE_DOCK      = None;
Atom ATOM_NET_WM_WINDOW_TYPE_TOOLBAR   = None;
Atom ATOM_NET_WM_WINDOW_TYPE_MENU      = None;
Atom ATOM_NET_WM_WINDOW_TYPE_DIALOG    = None;
Atom ATOM_NET_WM_WINDOW_TYPE_NORMAL    = None;

Atom ATOM_NET_WM_STATE                 = None;
Atom ATOM_NET_WM_STATE_MODAL           = None;
Atom ATOM_NET_WM_STATE_STICKY          = None;
Atom ATOM_NET_WM_STATE_MAXIMIZED_VERT  = None;
/* HORIZ or HORZ ? */
Atom ATOM_NET_WM_STATE_MAXIMIZED_HORIZ = None;
Atom ATOM_NET_WM_STATE_MAXIMIZED_HORZ  = None;
Atom ATOM_NET_WM_STATE_SHADED          = None;
Atom ATOM_NET_WM_STATE_SKIP_TASKBAR    = None;
Atom ATOM_NET_WM_STATE_SKIP_PAGER      = None;
/* _NET_WM_STATE_ADD, _NET_WM_STATE_REMOVE, _NET_WM_STATE_TOGGLE  are in net.h */

/* the next one is not in the spec but KDE use it */
Atom ATOM_NET_WM_STATE_STAYS_ON_TOP    = None;

Atom ATOM_NET_WM_STRUT                 = None;
Atom ATOM_NET_WM_ICON_GEOMETRY         = None;
Atom ATOM_NET_WM_ICON                  = None;
Atom ATOM_NET_WM_PID                   = None;
Atom ATOM_NET_WM_HANDLED_ICON          = None;

/* Window manager protocol */
Atom ATOM_NET_WM_PING = None;
Atom ATOM_WM_DELETE_WINDOW = None;
Atom ATOM_WM_PROTOCOLS = None;

/* some classical atoms */
/* Needed for the "name" atom */
Atom XA_UTF8_STRING = None;
Atom ATOM_WM_STATE = None;

/* KDE specific atoms: */
Atom ATOM_KDE_NET_SYSTEM_TRAY_WINDOWS       = None;
Atom ATOM_KDE_NET_WM_SYSTEM_TRAY_WINDOW_FOR = None;
Atom ATOM_KDE_NET_WM_FRAME_STRUT = None;
/* not yet used ... */
/* _KDE_WM_CHANGE_STATE, what is this? */
/* _KDE_NET_WM_WINDOW_TYPE_OVERRIDE,  what is this? */

/* FvwmNetHints atoms
 * CARDINAL =  MiniIconMaxSize which say that FvwmNetHints had set a (mini) 
 * icon */
Atom ATOM_FVWM_NET_SET_MINI_ICON = None;
/* has above for icon (replace <= by >) */
Atom ATOM_FVWM_NET_SET_ICON = None;
/* to manage AcceptStaysOnTop dynamically */
Atom ATOM_FVWM_NET_STAYS_ON_TOP = None;

/* Config (see also FvwmNetHints.c) */
Bool PagerSkipListUseSkipList = True;
Bool SkipListUsePagerSkipList = False;
Bool ForceSkipList = False;
Bool ForcePagerList = False;

Bool AcceptDesktopGeometry;
Bool AcceptStaysOnTop;
int StaysOnTopLayer;
int StaysPutLayer;

net_style DockNetStyle;
net_style ToolBarNetStyle;
net_style MenuNetStyle;
net_style DialogNetStyle;
net_style DesktopNetStyle;
/* this not a WM_TYPE it is WM_STATE ! */
net_style ModalNetStyle;

char *NetActivateWindow = NULL;
char *NetCloseWindow = NULL;

Bool UseStrutMaximize;
Bool UseDynamicStrutMaximize;
Bool UseStrutPlacement;
Bool UseExtendedName;
Bool StrutOverride;
unsigned DefaultHorizMaximize;
unsigned DefaultVertMaximize;

int MiniIconSize[2];
unsigned MiniIconMaxSize;

int PingId = 1;

/* extern variables */
extern char *MyName;
extern int fd[2];

extern ScreenInfo Scr;
extern Display *dpy;
extern NetWindow *Start;
extern NetWindow *FocusWin;
extern WindowList Stacking;
extern long NumberOfDesktops;
extern unsigned DoSendMiniIcon;

/* local functions */
Bool NET_CheckPingProto(Window w);
void NET_PingPrintInfo(NetWindow *netwin,pid_t pid, Window pid_win);
void NET_GetStrut(int *left, int *right, int *top, int *bottom, int desk, 
		  int dyn);
Window NET_GetPid(Window w, pid_t *pid);
void NET_SetNumberOfDesktops(void);
void NET_SetDesktopWorkArea(void);
void NET_AtomsInit(void);
int SetUpMwmHints(Window w, net_style style, CARD32 functions);
void SetUpWMHintsFocus(Window w, net_style style);
void CloseOldFvwmNetHints(void);
void NET_SetKdeSysTrayList(void);
void NET_GetWindowSystemTray(void);

/* ***************************************************************************
 * XAtom Utils
 * *************************************************************************** */

void *NET_AtomGet(Window win, Atom to_get, Atom type, int *size)
{
  unsigned char	*retval;
  Atom	type_ret;
  unsigned long	 bytes_after, num_ret;
  long length;
  int  format_ret;
  void *data;

  retval = NULL;
  length = 0x7fffffff;
  XGetWindowProperty(dpy, win, to_get, 0,
		     length,
		     False, type,
		     &type_ret,
		     &format_ret,
		     &num_ret,
		     &bytes_after,
		     &retval);

  if ((retval) && (num_ret > 0) && (format_ret > 0))
  {
    data = safemalloc(num_ret * (format_ret >> 3));
    if (data)
      memcpy(data, retval, num_ret * (format_ret >> 3));
    XFree(retval);
    *size = num_ret * (format_ret >> 3);
    return data;
  }
  return NULL;
}

/* ***************************************************************************
 * Handling the Property Notify
 * *************************************************************************** */
void NET_ProcessPropertyNotify(XEvent *Event)
{
  Window w = Event->xclient.window;
  NetWindow *netwin = GetNetWindow(w);

  if (netwin == NULL)
    return;

  /*fprintf(stderr,"************ Property Notify **************(%li,0x%lx)\n",
    Event->xproperty.atom, w);*/

  if (Event->xproperty.atom == ATOM_NET_WM_ICON) 
  {
#if HAVE_AN_IMLIB
    if (DoSendMiniIcon != DO_NOT_SEND_NET_ICON &&
	(DO_HAS_NET_MINI_ICON(netwin) || DO_HAS_SIZE_NET_MINI_ICON(netwin)))
      {
	/*fprintf(stderr,"************ Rebuild NET Icon **************\n");*/
	if (BuildMiniIcon(netwin, NULL, False))
	  SendMiniIcon(netwin->w);
      }
#endif
    return;
  }
  else if (Event->xproperty.atom == ATOM_NET_WM_ICON_GEOMETRY)
  {
    /*fprintf(stderr,"************ ATOM_NET_WM_ICON_GEOMETRY\n");*/
  }
  else if (Event->xproperty.atom == ATOM_NET_WM_STRUT)
  {
    int left,right,top,bottom;
    NET_GetWindowStrut(&left,&right,&top,&bottom,netwin->w);
    
    if (left != netwin->strut_w.left || netwin->strut_w.right != right ||
	netwin->strut_w.top != top || netwin->strut_w.bottom != bottom)
    {
      SET_DO_HAS_STRUT(netwin,True);
      if (left > 0 || right > 0 || top > 0 || bottom > 0)
      {
	netwin->strut_w.left = left;
	netwin->strut_w.right = right;
	netwin->strut_w.top = top;
	netwin->strut_w.bottom = bottom;
	NET_SetDesktopWorkArea();
      }
    }
    if (left != netwin->dyn_strut_w.left || netwin->dyn_strut_w.right != right ||
	netwin->dyn_strut_w.top != top || netwin->dyn_strut_w.bottom != bottom)
    {
      	netwin->dyn_strut_w.left = left;
	netwin->dyn_strut_w.right = right;
	netwin->dyn_strut_w.top = top;
	netwin->dyn_strut_w.bottom = bottom;
      if (UseDynamicStrutMaximize)
	NET_UpdateMaximizedWindows();
    }
  }
  else if (Event->xproperty.atom == ATOM_WM_STATE)
  {
    /* bad hack ... to fix kicker system tray, when kicker restart*/
    if (netwin->window_name != NULL && netwin->res_name != NULL &&
	strcmp(netwin->window_name,"kicker") == 0 &&
	strcmp(netwin->res_name,"Panel") == 0) {
      FreeList(&KdeSysTray);
      InitList(&KdeSysTray);
    }
  }
  else if (Event->xproperty.atom == ATOM_NET_WM_NAME)
  {
    if (UseExtendedName)
      NET_SetVisibleName(netwin,False);
  }
  else if (Event->xproperty.atom == ATOM_NET_WM_ICON_NAME)
  {
    if (UseExtendedName)
      NET_SetVisibleName(netwin,True);
  }
}

/* ***************************************************************************
 * Handling the Client Message
 * *************************************************************************** */
void NET_ProcessClientMessage(XEvent *Event)
{
  char command[256];
  Window w = Event->xclient.window;
  NetWindow *netwin = GetNetWindow(w);

  /* First Handle messages not assosiated to a window */

  if (Event->xclient.message_type == ATOM_WM_PROTOCOLS && w == Scr.Net_w)
  {
    if (Event->xclient.data.l[0] == ATOM_WM_DELETE_WINDOW)
    {
      exit(0);
    }
    else if (Event->xclient.data.l[0] == ATOM_NET_WM_PING)
    {
      /* really I do not see why we should get such message :o)*/
      send_clientmessage(dpy, Scr.Root, ATOM_NET_WM_PING,
			 Event->xclient.data.l[1]);
    }
  }
  else if (Event->xclient.message_type == ATOM_WM_PROTOCOLS)
  {
    if (Event->xclient.data.l[0] == ATOM_NET_WM_PING) {
      if (IsPingItemInfo(&PingList, Event->xclient.data.l[1]))
	  fprintf(stderr, "%s: Get a ping answer with id: %li\n",
		  MyName,Event->xclient.data.l[1]);
      DeletePingItem(&PingList,Event->xclient.data.l[1]);
    }
  }
  else if (Event->xclient.message_type == ATOM_NET_CURRENT_DESKTOP)
  {
    int desk = Event->xclient.data.l[0];

    if (desk >= 0) {
      sprintf(command,"GotoDesk 0 %i",desk);
      SendText(fd,command,0);
    }
    /* We do not need to: NET_SetNumberOfDesktops() because FVWM will sent to
       us a M_DESK message, the same remark applies below */
  }

  else if (Event->xclient.message_type == ATOM_NET_NUMBER_OF_DESKTOPS)
  {
    /* not a lot of sinification for FVWM */
    int d = Event->xclient.data.l[0];

    if (d > 0 && d <= MAX_NUMBER_OF_DESKTOPS) {
      NumberOfDesktops = d;
      NET_SetNumberOfDesktops();
    }
  }

  else if (Event->xclient.message_type == ATOM_NET_DESKTOP_GEOMETRY)
  {
    int x = Event->xclient.data.l[0];
    int y = Event->xclient.data.l[1];

    if (AcceptDesktopGeometry) {
      sprintf(command,"DesktopSize %ix%i",x,y);
      SendFvwmPipe(fd,command, w);
    }
  }

  else if (Event->xclient.message_type == ATOM_NET_DESKTOP_VIEWPORT)
  {
    int x = Event->xclient.data.l[0];
    int y = Event->xclient.data.l[1];

    sprintf(command,"Scroll %ip %ip",x,y);
    SendFvwmPipe(fd,command, w);
  }
  /* else if (netwin == NULL) {
    fprintf(stderr,"UnKNOWN NET Client message (no window): %i,%i,%i,%i,%i\n",
	    Event->xclient.message_type,Event->xclient.data.l[0],
	    Event->xclient.data.l[1],Event->xclient.data.l[2],
	    Event->xclient.data.l[3],Event->xclient.data.l[4]);
	    }*/

  /***********************************
   *  Now we need a window
   ***********************************/

  if (netwin == NULL) {
    return;
  }

  /* Some times kdesktop want to be activated, for us a desktop window
   * is more or less equivalent to the root window */ 
  if (Event->xclient.message_type == ATOM_NET_ACTIVE_WINDOW)
  {
    /* When kdesktop want to popup a menu (overide redirect) it ask to be 
     * activated (good idea). But for us a desktop window is more or less 
     * equivalent to the root window. Any way we should do somethings! */
    if (netwin->net_window_type == ATOM_NET_WM_WINDOW_TYPE_DESKTOP) {
      /* we must ?? give the focus to a dummy window */
      sprintf(command,"SendToModule fvwm2 UnFocusWindow");
      SendFvwmPipe(fd,command,w);
    } else {
      sprintf(command,"%s", NetActivateWindow);
      SendFvwmPipe(fd,command,w);
    }
  }

  else if (Event->xclient.message_type == ATOM_NET_CLOSE_WINDOW)
  {
    sprintf(command,NetCloseWindow);
    SendText(fd,command,w);
  }

  else if (Event->xclient.message_type == ATOM_NET_WM_DESKTOP)
  {
    unsigned long d = (unsigned long)Event->xclient.data.l[0];

    /* the spec says that if d = 0xFFFFFFFF then we have to Stick the window 
    *  however KDE use 0xFFFFFFFE :o) */ 
    if (d == 0xFFFFFFFE || d == 0xFFFFFFFF) {
      sprintf(command,"Silent Stick On");
      SendText(fd,command,w);
    }
    else if (d >= 0) {
      if (IS_STICKY(netwin)) {
	sprintf(command,"Silent Stick Off");
	SendText(fd,command,w);
      }
      sprintf(command,"MoveToDesk 0 %ld",d);
      SendText(fd,command,w);
    }
  }

  else if (Event->xclient.message_type == ATOM_NET_WM_MOVERESIZE)
  {
    /* I am not sure to have understand this message ... */
    int dir = Event->xclient.data.l[2];
    char action[10];
    int x_warp = 0;
    int y_warp = 0;

    sprintf(action,"Resize");

    /* our default is _NET_WM_MOVERESIZE_SIZE_TOPLEFT */
    if (dir == _NET_WM_MOVERESIZE_SIZE_TOP)
      x_warp = 50;
    else if (dir == _NET_WM_MOVERESIZE_SIZE_TOPRIGHT)
      x_warp = 100;
    else if (dir == _NET_WM_MOVERESIZE_SIZE_RIGHT) {
      x_warp = 100; y_warp = 50;
    }
    else if (dir == _NET_WM_MOVERESIZE_SIZE_BOTTOMRIGHT) {
      x_warp = 100; y_warp = 100;
    }
    else if (dir == _NET_WM_MOVERESIZE_SIZE_BOTTOM) {
      x_warp = 50; y_warp = 100;
    }
    else if (dir == _NET_WM_MOVERESIZE_SIZE_BOTTOMLEFT)
      y_warp = 100;
    else if (dir == _NET_WM_MOVERESIZE_SIZE_LEFT)
      y_warp = 50;
    else if (dir == _NET_WM_MOVERESIZE_MOVE) {
      x_warp = 50; y_warp = 50; /* why not */
      sprintf(action,"Move");
    }
    sprintf(command,"WarpToWindow %i %i, %s",x_warp,y_warp,action);
    SendFvwmPipe(fd,command,w);
  }

  else if (Event->xclient.message_type == ATOM_NET_WM_STATE)
  {
    Window w = Event->xclient.window;
    int action = Event->xclient.data.l[0];
    Atom first_prop = Event->xclient.data.l[1];
    Atom second_prop = Event->xclient.data.l[2];
    int max_horiz = 0;
    int max_vert = 0;
    char ct[7];
    Bool send_cmd = False;

    if (action == _NET_WM_STATE_REMOVE) {
      sprintf(ct,"Off");
    }
    else if (action == _NET_WM_STATE_ADD) {
      sprintf(ct,"On");
    }
    else {
      sprintf(ct,"toggle");
    }

    /* first prop */
    if (first_prop == ATOM_NET_WM_STATE_MODAL) {
      /* IMHO: problems in the spec */
    }
    else if (first_prop == ATOM_NET_WM_STATE_STICKY) {
      sprintf(command,"Silent Stick %s",ct);
      SendText(fd,command,w);
    }
    else if (first_prop == ATOM_NET_WM_STATE_MAXIMIZED_VERT) {
      max_vert = DefaultVertMaximize;
    }
    else if (first_prop == ATOM_NET_WM_STATE_MAXIMIZED_HORIZ ||
	     first_prop == ATOM_NET_WM_STATE_MAXIMIZED_HORZ) {
      max_horiz = DefaultHorizMaximize;
    }
    else if (first_prop == ATOM_NET_WM_STATE_SHADED) {
      sprintf(command,"Silent WindowShade %s",ct);
      SendText(fd,command,w);
    }
    else if (first_prop == ATOM_NET_WM_STATE_SKIP_TASKBAR) {
      if ((action == _NET_WM_STATE_TOGGLE && !DO_SKIP_WINDOW_LIST(netwin))||
	  action == _NET_WM_STATE_ADD) {
	send_cmd = True;
	sprintf(command,"Silent Style $n WindowListSkip");
      }
      else {
	send_cmd = True;
	sprintf(command,"Silent Style $n WindowListHit");
      }
      if (send_cmd)
	SendText(fd,command,w);
    }
    /* NOT SUPPORTED BY FVWM */
    else if (first_prop == ATOM_NET_WM_STATE_SKIP_PAGER) {
      if ((action == _NET_WM_STATE_TOGGLE && 
	   !DO_SKIP_PAGER_WINDOW_LIST(netwin))|| action == _NET_WM_STATE_ADD) {
	UpdateSkipPager(netwin->w,True);
	if (SkipListUsePagerSkipList && !DO_SKIP_WINDOW_LIST(netwin)) {
	  sprintf(command,"Silent Style $n WindowListSkip");
	  send_cmd = True;
	}
      }
      else {
	UpdateSkipPager(netwin->w,False);
	if (SkipListUsePagerSkipList && DO_SKIP_WINDOW_LIST(netwin)) {
	  send_cmd = True;
	  sprintf(command,"Silent Style $n WindowListHit");
	}
      }
      if (send_cmd)
	SendText(fd,command,w);
    }
    /* Problem: not in the spec and need to get the default layer! */ 
    else if (first_prop == ATOM_NET_WM_STATE_STAYS_ON_TOP) {
      if (action == _NET_WM_STATE_ADD || (action == _NET_WM_STATE_TOGGLE &&
					  netwin->layer < StaysOnTopLayer)) {
	sprintf(command,"Silent Layer 0 %i",StaysOnTopLayer);
	SendText(fd,command,w);
      }
      else {
	sprintf(command,"Silent Layer 0 %i",StaysPutLayer);
	SendText(fd,command,w);
      }
    }

    /* second action */
    if (second_prop == ATOM_NET_WM_STATE_MAXIMIZED_VERT) {
      max_vert = DefaultVertMaximize;
    }
    else if (second_prop == ATOM_NET_WM_STATE_MAXIMIZED_HORIZ ||
	     second_prop == ATOM_NET_WM_STATE_MAXIMIZED_HORZ) {
      max_horiz = DefaultHorizMaximize;
    }
    else if (second_prop == 0)
    {
      /* nothing to do */
    }

    if (max_vert != 0 || max_horiz != 0) {
      NET_StrutMaximize(max_horiz, max_vert, action, netwin);
    }
  }
  /*else {
    fprintf(stderr,"UNKNOWN NET Client message (Ox%lx): %i\n",
	    w,Event->xclient.message_type);
	    }*/
  
}

/* ***************************************************************************
 * Strut Maximizing: maximizing with respect to the (dynamic) WorkArea.
 * *************************************************************************** */
void NET_StrutMaximize(int horiz, int vert, int action, NetWindow *netwin)
{

  char command[256];
  float h,v;
  int left = 0;
  int right = 0;
  int top = 0;
  int bottom = 0;
  
  if (action == _NET_WM_STATE_REMOVE || (IS_MAXIMIZED(netwin) &&
					 action == _NET_WM_STATE_TOGGLE))
  {
    SendText(fd,"Maximize Off",netwin->w);
    UpdateStrutMaximize(netwin->w,0,0,0,0,0,0);
    return;
  }

  if (!UseStrutMaximize)
  {
    sprintf(command,"Silent Maximize On %i %i",horiz,vert);
    SendText(fd,command,netwin->w);
    return;
  }

  if (!UseDynamicStrutMaximize)
  {
     /* we should maximize in the work space */
    NET_GetStrut(&left, &right, &top, &bottom, netwin->desk,0);
  }
  else
  {
     /* we should maximize in the dynamic work space */
    NET_GetStrut(&left, &right, &top, &bottom, netwin->desk,1);
  }

  h = horiz * ( (float)(Scr.MyDisplayWidth - left - right) / 
		(float)(Scr.MyDisplayWidth));
  v = vert *  ( (float)(Scr.MyDisplayHeight - top - bottom) / 
		(float)(Scr.MyDisplayHeight));

  h = (int) h;
  v = (int) v;
  
  if (netwin->s_max.v == (int) v && netwin->s_max.h == (int) h &&
      netwin->s_max.x == left && netwin->s_max.y == top)
  {
    return;
  }
  
  sprintf(command,"Silent Maximize On %i %i",(int)h,(int)v);
  SendText(fd,command,netwin->w);
  if (left !=0 || top !=0) {
    int l = (h == 0)? 0:left;
    int t = (v == 0)? 0:top;
    sprintf(command,"Silent Move w%ip w%ip",l,t);
    SendText(fd,command,netwin->w);
  }
  UpdateStrutMaximize(netwin->w, horiz, vert, (int)h, (int)v, left, top);
}

/* ***************************************************************************
 * Remaximize the windows (for dynamic maximizing)
 * *************************************************************************** */
void  NET_UpdateMaximizedWindows(void)
{
  NetWindow *t;

  t = Start;

  while(t!= NULL)
  {
    if (t->s_max.vert !=0 || t->s_max.horiz != 0)
      NET_StrutMaximize(t->s_max.horiz, t->s_max.vert, _NET_WM_STATE_ADD, t);
    t = t->next;
  }
}

/* ***************************************************************************
 * NET Handle window type
 * *************************************************************************** */
Atom NET_HandleWindowType(NetWindow *netwin)
{
  unsigned int size = 0;
  Atom *val;
  unsigned int i = 0;
  char command[256];
  int do_redecorate = 0;
  int layer;
  Atom ret = ATOM_NET_WM_WINDOW_TYPE_NORMAL;

  val = NET_AtomGet(netwin->w, ATOM_NET_WM_WINDOW_TYPE, XA_ATOM, &size);

  if (val == NULL)
    return ret;

  ret = val[0];
  /* we support only one window type */
  /* for(i = 0; i < (size / (sizeof(Atom))); i++) */ 
  {
    layer = -1;
    if (val[i] == ATOM_NET_WM_WINDOW_TYPE_DOCK) 
    {
      if (DockNetStyle.layer > 0)
	layer = DockNetStyle.layer;
      do_redecorate = SetUpMwmHints(netwin->w,DockNetStyle,MWM_FUNC_ALL);
      SetUpWMHintsFocus(netwin->w,DockNetStyle);
      sprintf(command, "SendToModule fvwm2 NET_WindowConfig %u %u %u %u %i %u",
	      0,DockNetStyle.is_sticky,
	      DockNetStyle.do_circulate_skip,
	      DockNetStyle.do_window_list_skip,
	      layer,do_redecorate);
      SendText(fd,command,netwin->w);
    }

    else if (val[i] == ATOM_NET_WM_WINDOW_TYPE_TOOLBAR)
    {
      if (ToolBarNetStyle.layer > 0)
	layer = ToolBarNetStyle.layer;
      do_redecorate = SetUpMwmHints(netwin->w,ToolBarNetStyle,MWM_FUNC_ALL);
      SetUpWMHintsFocus(netwin->w,ToolBarNetStyle);
      sprintf(command, "SendToModule fvwm2 NET_WindowConfig %u %u %u %u %i %u",
	      0,ToolBarNetStyle.is_sticky,
	      ToolBarNetStyle.do_circulate_skip,
	      ToolBarNetStyle.do_window_list_skip,
	      layer,do_redecorate);
      SendText(fd,command,netwin->w);
    }

    else if (val[i] == ATOM_NET_WM_WINDOW_TYPE_MENU)
    {
      if (MenuNetStyle.layer > 0)
	layer = MenuNetStyle.layer;
      do_redecorate = SetUpMwmHints(netwin->w,MenuNetStyle,0);
      SetUpWMHintsFocus(netwin->w,MenuNetStyle);
      sprintf(command, "SendToModule fvwm2 NET_WindowConfig %u %u %u %u %i %u",
	      0,MenuNetStyle.is_sticky,
	      MenuNetStyle.do_circulate_skip,
	      MenuNetStyle.do_window_list_skip,
	      layer,do_redecorate);
      SendText(fd,command,netwin->w);
    }

    else if (val[i] == ATOM_NET_WM_WINDOW_TYPE_DIALOG)
    {
      if (DialogNetStyle.layer > 0)
	layer = DialogNetStyle.layer;
      do_redecorate = SetUpMwmHints(netwin->w,DialogNetStyle,MWM_FUNC_ALL);
      SetUpWMHintsFocus(netwin->w,DialogNetStyle);
      sprintf(command, "SendToModule fvwm2 NET_WindowConfig %i %i %i %i %i %i",
	      0,DialogNetStyle.is_sticky,
	      DialogNetStyle.do_circulate_skip,
	      DialogNetStyle.do_window_list_skip,
	      layer,do_redecorate);
      SendText(fd,command,netwin->w);
    }

    else if (val[i] == ATOM_NET_WM_WINDOW_TYPE_NORMAL)
    {
    }
      
    else if (val[i] == ATOM_NET_WM_WINDOW_TYPE_DESKTOP) 
    {
      layer = LAYER_NET_WM_WINDOW_TYPE_DESKTOP;
      do_redecorate = SetUpMwmHints(netwin->w,DesktopNetStyle,MWM_FUNC_ALL);
      SetUpWMHintsFocus(netwin->w,DesktopNetStyle);
      sprintf(command, "SendToModule fvwm2 NET_WindowConfig %u %u %u %u %i %u",
	      1,DesktopNetStyle.is_sticky,
	      DesktopNetStyle.do_circulate_skip,
	      DesktopNetStyle.do_window_list_skip,layer,do_redecorate);
      SendText(fd,command,netwin->w);
    }
  } /* for */
  
  free(val);
  return ret;
}

/* **************************************************************************
 *
 * ************************************************************************** */
void NET_ReconfigureWindowType(NetWindow *netwin)
{
  char command[256];
  net_style st;
  int layer = -1;

  if (netwin->net_window_type == ATOM_NET_WM_WINDOW_TYPE_DOCK)
    st = DockNetStyle;
  else if (netwin->net_window_type == ATOM_NET_WM_WINDOW_TYPE_TOOLBAR)
    st = ToolBarNetStyle;
  else if (netwin->net_window_type == ATOM_NET_WM_WINDOW_TYPE_MENU)
    st = MenuNetStyle;
  else if (IS_NET_MODAL(netwin))
    st = ModalNetStyle;
  else if (netwin->net_window_type == ATOM_NET_WM_WINDOW_TYPE_DIALOG)
    st = DialogNetStyle;
  else if (netwin->net_window_type == ATOM_NET_WM_WINDOW_TYPE_DESKTOP) {
    st = DesktopNetStyle;
    layer = LAYER_NET_WM_WINDOW_TYPE_DESKTOP;
  }
  else
    return;

  if (st.layer > 0 &&
       netwin->net_window_type != ATOM_NET_WM_WINDOW_TYPE_DESKTOP)
    layer = st.layer;

  if ((IS_STICKY(netwin) != st.is_sticky) ||
      (DO_SKIP_WINDOW_LIST(netwin) != st.do_window_list_skip) ||
      (DO_SKIP_CIRCULATE(netwin) != st.do_circulate_skip) ||
      (netwin->layer != layer && layer != -1))
  {
    sprintf(command, "SendToModule fvwm2 NET_WindowConfig %u %u %u %u %i %u",
	    0,st.is_sticky,st.do_circulate_skip,st.do_window_list_skip,layer,0);
    SendText(fd,command,netwin->w);
  }
}

/* **************************************************************************
 * NET Handle WM State, this may be merged with coresponding part of 
 * NET_ProcessClientMessage
 * ************************************************************************** */
void NET_HandleWMState(NetWindow *netwin,int *do_skip_pager,
		       int *is_net_stays_on_top, int *is_modal)
{
  unsigned int size = 0;
  Atom *val;
  unsigned int i;
  char command[256];
  int max_vert = 0;
  int max_horiz = 0;
  unsigned int s = 0;
  CARD32 *fvwm_sot;

  *is_modal = 0;
  *is_net_stays_on_top = 0;
  *do_skip_pager = 0;

  val = NET_AtomGet(netwin->w, ATOM_NET_WM_STATE, XA_ATOM, &size);
  fvwm_sot =
    NET_AtomGet(netwin->w, ATOM_FVWM_NET_STAYS_ON_TOP, XA_CARDINAL, &s);

  if (fvwm_sot != NULL && fvwm_sot[0] == STAYS_ON_TOP_SET_BY_APPLICATION)
    *is_net_stays_on_top = 1;

  if (val == NULL) {
    if (fvwm_sot != NULL)
      free(fvwm_sot);
    return;
  }

  for(i = 0; i < (size / (sizeof(Atom))); i++) 
  {

    if (val[i] == ATOM_NET_WM_STATE_MODAL)
    {
      int layer = -1;
      int do_redecorate;

      *is_modal = 1;
      if (ModalNetStyle.layer > 0)
	layer = ModalNetStyle.layer;
      do_redecorate = SetUpMwmHints(netwin->w,ModalNetStyle,MWM_FUNC_ALL);
      SetUpWMHintsFocus(netwin->w,ModalNetStyle);
      sprintf(command, "SendToModule fvwm2 NET_WindowConfig %i %i %i %i %i %i",
	      0,ModalNetStyle.is_sticky,
	      ModalNetStyle.do_circulate_skip,
	      ModalNetStyle.do_window_list_skip,
	      layer,do_redecorate);
      SendText(fd,command,netwin->w);
    }
  
    else if (val[i] == ATOM_NET_WM_STATE_STICKY) 
    {
      if (!IS_STICKY(netwin)) {
	sprintf(command,"Silent Stick On");
	SendText(fd,command,netwin->w);
      }
    }
      
    else if (val[i] == ATOM_NET_WM_STATE_MAXIMIZED_VERT)
    {
      max_vert = 100;
    }

    else if (val[i] == ATOM_NET_WM_STATE_MAXIMIZED_HORIZ ||
	     val[i] == ATOM_NET_WM_STATE_MAXIMIZED_HORZ) 
    {
      max_horiz = 100;
    }

    else if (val[i] == ATOM_NET_WM_STATE_SHADED)
    {
      if (!IS_SHADED(netwin)) {
	sprintf(command,"Silent WindowShade On");
	SendText(fd,command,netwin->w);
      }
    }

    else if (val[i] == ATOM_NET_WM_STATE_SKIP_TASKBAR)
    {
      if (!DO_SKIP_WINDOW_LIST(netwin)) {
	sprintf(command,"Silent Style $n WindowListSkip");
	SendText(fd,command,netwin->w);
      }
      if (PagerSkipListUseSkipList)
	*do_skip_pager = 1;
    }

    else if (val[i] == ATOM_NET_WM_STATE_SKIP_PAGER) 
    {
      /* What to do ? */
      if (SkipListUsePagerSkipList) {
	if (!DO_SKIP_WINDOW_LIST(netwin)) {
	  sprintf(command,"Silent Style $n WindowListSkip");
	  SendText(fd,command,netwin->w);
	}
      }
      *do_skip_pager = 1;
    }
      
    else if (val[i] == ATOM_NET_WM_STATE_STAYS_ON_TOP)
    {
      if (fvwm_sot == NULL || fvwm_sot[0] == STAYS_ON_TOP_SET_BY_APPLICATION) 
	*is_net_stays_on_top = 1;
      if (AcceptStaysOnTop && StaysOnTopLayer > 0) {
	if (netwin->layer < StaysOnTopLayer) {
	  sprintf(command,"Silent Layer 0 %i", StaysOnTopLayer);
	  SendText(fd,command,netwin->w);
	}
      }
    }
#ifdef NET_HINTS_WARNING_MSG
    else
    {
      fprintf(stderr,"%s [Warning]: Unknown WM_STATE %lu\n",MyName, val[i]); }
#endif
  } /* for */

  if (max_vert != 0 || max_horiz != 0) {
    if (!IS_MAXIMIZED(netwin)) {
      /* sprintf(command,"Maximize %i %i", max_vert, max_horiz);
       * SendText(fd,command,netwin->w); */
      NET_StrutMaximize(max_horiz, max_vert, _NET_WM_STATE_ADD, netwin);
    }
  }
  if (fvwm_sot != NULL)
    free(fvwm_sot);
  free(val);
}

void NET_HandleKdeSysTray(Window w)
{
  unsigned int size = 0;
  Atom *val;

  val = NET_AtomGet(w, ATOM_KDE_NET_WM_SYSTEM_TRAY_WINDOW_FOR,XA_WINDOW, &size);

  if (val == NULL)
    return;

  /* problem: How/When to delete the elements of this list? */
  /* I've done a bad hack in ProcessPrpertyNotify */
  AddItem(&KdeSysTray,w);
  if (Started)
    NET_SetKdeSysTrayList();
}

/* ***************************************************************************
 * PING protocol functions
 * *************************************************************************** */

/* ***************************************************************************
 * PING implementation
 * *************************************************************************** */
void NET_PingWindow(NetWindow *netwin, int action, int time_out)
{
  pid_t pid;
  Window pid_w;
  Bool ping_proto = False;
  Time timestamp;

  pid_w = NET_GetPid(netwin->w,&pid);
  PingId++;
  if (PingId > 32000)
    PingId = 1;

  if (action == PING_ACTION_INFO)
    NET_PingPrintInfo(netwin, pid, pid_w);

  /* It will maybe better to ping an other window */
  ping_proto = NET_CheckPingProto(netwin->w);

  if ((action != PING_ACTION_DELETE && action != PING_ACTION_PING_TERM &&
       action != PING_ACTION_PING_KILL && action != PING_ACTION_INFO)
      || ping_proto) {
    time_t t = time(NULL);
    AddPingItem(&PingList,netwin->w,PingId,action,t,time_out,pid);
    SetAlarm();
  }

  if (ping_proto) {
    timestamp = (Time)PingId;
    send_clientmessage(dpy,netwin->w,ATOM_NET_WM_PING,timestamp);
    if (action == PING_ACTION_INFO) {
      fprintf(stderr,"\n    -- Send a PING to this window with id: %ld --\n",
	      timestamp);
    }
  }
  else if (action == PING_ACTION_INFO)
    fprintf(stderr,"\n    This window does not support the PING Protocol\n");

  if (action == PING_ACTION_DELETE)
    SendInfo(fd,"Silent Delete",netwin->w);
  else if (action == PING_ACTION_CLOSE)
    SendInfo(fd,"Silent Close",netwin->w);
  else if (action == PING_ACTION_DESTROY)
    SendInfo(fd,"Silent Destroy",netwin->w);
}

/* ***************************************************************************
 * check if the PING protocol is implemented
 * *************************************************************************** */
Bool NET_CheckPingProto(Window w)
{
  Atom *val;
  unsigned int size = 0;
  int i;
  Bool ping_proto = False;
  
  val = NET_AtomGet(w, ATOM_WM_PROTOCOLS, XA_ATOM, &size);

  if (val != NULL) {
    for(i = 0; i < (size / (sizeof(Atom))); i++)
    {
      if (val[i] == ATOM_NET_WM_PING)
	ping_proto = True;
    }
    free(val);
  }
  return ping_proto;
}

/* ***************************************************************************
 * Print Some NET information
 * *************************************************************************** */
void NET_PingPrintInfo(NetWindow *netwin, pid_t pid, Window pid_w)
{

  fprintf(stderr,"\n%s: Ping Window 0x%lx\n\
    Window Name: %s\n\
    Icon Name  : %s\n\
    Res Name   : %s\n\
    Res Class  : %s\n\
    **** NET INFO ****
    WINDOW TYPE: ",
	  MyName,
	  netwin->w,
	  (netwin->window_name != NULL)? netwin->window_name:"not set",
	  (netwin->icon_name != NULL)? netwin->icon_name:"not set",
	  (netwin->res_name != NULL)? netwin->res_name:"not set",
	  (netwin->res_class != NULL)? netwin->res_class:"not set");

  if (netwin->net_window_type == ATOM_NET_WM_WINDOW_TYPE_DESKTOP)
    fprintf(stderr,"DESKTOP\n");
  else if (netwin->net_window_type == ATOM_NET_WM_WINDOW_TYPE_DOCK)
    fprintf(stderr,"DOCK\n");
  else if (netwin->net_window_type == ATOM_NET_WM_WINDOW_TYPE_TOOLBAR)
    fprintf(stderr,"TOOLBAR\n");
  else if (netwin->net_window_type == ATOM_NET_WM_WINDOW_TYPE_MENU)
    fprintf(stderr,"MENU\n");
  else if (netwin->net_window_type == ATOM_NET_WM_WINDOW_TYPE_DIALOG)
    fprintf(stderr,"DIALOG\n");
  else if (netwin->net_window_type == ATOM_NET_WM_WINDOW_TYPE_NORMAL)
    fprintf(stderr,"NORMAL\n");
  else
    fprintf(stderr,"NORMAL (not set)\n");

  fprintf(stderr,"\
    STRUT      : %i %i %i %i\n\
    DYN STRUT  : %i %i %i %i\n\
    SKIP_PAGER : %s\n\
    IS MODAL   : %s\n\
    ON TOP     : %s (%i)\n",
	  netwin->strut_w.left,netwin->strut_w.right,
	  netwin->strut_w.top,netwin->strut_w.bottom,
	  netwin->dyn_strut_w.left,netwin->dyn_strut_w.right,
	  netwin->dyn_strut_w.top,netwin->dyn_strut_w.bottom,
	  (DO_SKIP_PAGER_WINDOW_LIST(netwin))? "Yes":"No",
	  (IS_NET_MODAL(netwin))? "Yes":"No",
	  (IS_NET_STAYS_ON_TOP(netwin))? 
	  ((AcceptStaysOnTop)? "Yes Accepted":"Yes Rejected"):"No",
	  netwin->layer);

  fprintf(stderr,"\
    MINI ICON  : %i %i\n\
    ICON       : %i %i\n",
	  netwin->my_flags.net_mini_icon_state,
	  netwin->my_flags.net_mini_icon_mode,
	  netwin->my_flags.net_icon_state,
	  netwin->my_flags.net_icon_mode);

  fprintf(stderr,"\
    PID        : %i %s",pid, (pid)? "":"(not set)");
  if (pid && pid_w != netwin->w)
    fprintf(stderr," (for leader window 0x%lx)\n",pid_w);
  else
    fprintf(stderr,"\n");
}

/* ***************************************************************************
 * "Get" Function
 * *************************************************************************** */

/* ***************************************************************************
 * NET get the Pid
 * *************************************************************************** */
Window NET_GetPid(Window w, pid_t *pid)
{
  unsigned int size = 0;
  CARD32 *val;
  XWMHints *xwmh;
  Window pid_win = w;

  *pid = 0;
  val = NET_AtomGet(w, ATOM_NET_WM_PID, XA_CARDINAL, &size);

  /* with GTK+ all application have a leader window which is the "owner"
   * of the pid (and which does not support the WM_PING protocol). */
  if (val == NULL) {
    xwmh = XGetWMHints(dpy, w);
    if (xwmh !=NULL) {
      if (xwmh->flags & WindowGroupHint) {
	pid_win = xwmh->window_group;
	val = NET_AtomGet(pid_win, ATOM_NET_WM_PID, XA_CARDINAL, &size);
      }
      XFree(xwmh);
    }
  }

  if (val != NULL) {
    *pid = val[0];
    free(val);
  }
  return pid_win;
}

/* ***************************************************************************
 * Get the Strut of a window
 * *************************************************************************** */
void NET_GetWindowStrut(int *left, int *right, int *top, int *bottom, Window w)
{
  unsigned int size = 0;
  CARD32 *val;  

  *left = 0;
  *right = 0;
  *top = 0;
  *bottom = 0;

  val = NET_AtomGet(w, ATOM_NET_WM_STRUT, XA_CARDINAL, &size);

  if (val == NULL)
    return;

  if ((size / (sizeof(CARD32))) != 4) {
#ifdef NET_HINTS_WARNING_MSG
    fprintf(stderr,"%s [Warning]: Window 0x%lx has wrong STRUT value (%d)\n",
	    MyName,w,size / (sizeof(CARD32)));
#endif
    free(val);
    return;
  }
  
  *left   = val[0];
  *right  = val[1];
  *top    = val[2];
  *bottom = val[3];
  free(val);  
}

/* ***************************************************************************
 * Get the (dynamic) "Strut"/WorkArea for a given desk
 * *************************************************************************** */
void NET_GetStrut(int *left, int *right, int *top, int *bottom, int desk, 
		  int dyn)
{
  NetWindow *t;

  *left = 0;
  *right = 0;
  *top = 0;
  *bottom = 0;
  
  t = Start;

  while(t!= NULL && !StrutOverride)
  {
    if (DO_HAS_STRUT(t) && (t->desk == desk ||  IS_STICKY(t)))
    {
      if (dyn) {
	*left   = max(*left,t->dyn_strut_w.left);
	*right  = max(*right,t->dyn_strut_w.right);
	*top    = max(*top,t->dyn_strut_w.top);
	*bottom = max(*bottom,t->dyn_strut_w.bottom);
      }
      else
      {
	*left   = max(*left,t->strut_w.left);
	*right  = max(*right,t->strut_w.right);
	*top    = max(*top,t->strut_w.top);
	*bottom = max(*bottom,t->strut_w.bottom);
      }
    }
    t = t->next;
  }
  *left   = max(Scr.my_strut.left,*left);
  *right  = max(Scr.my_strut.right,*right);
  *top    = max(Scr.my_strut.top,*top);
  *bottom = max(Scr.my_strut.bottom,*bottom);
}

/* ***************************************************************************
 * Get the icon location (for animation)
 * There are problems in the spec here
 * *************************************************************************** */
void NET_GetWmIconGeometry(int *x, int *y, int *width, int *height, Window w)
{
  unsigned int size = 0;
  CARD32 *val;

  *x = -1;
  *y = -1;
  *width = -1;
  *height = -1;
  
  val = NET_AtomGet(w, ATOM_NET_WM_ICON_GEOMETRY, XA_CARDINAL, &size);

  if (val == NULL)
    return;

  if ((size / (sizeof(CARD32))) != 4) {
#ifdef NET_HINTS_WARNING_MSG
    fprintf(stderr,"%s [Warning]: Window 0x%lx has wrong WM ICON GEOMETRY(%d)\n",
	    MyName,w,size/(sizeof(CARD32)));
#endif
    free(val);
    return;
  }

  *x = val[0];
  *y = val[1];
  *width = val[2];
  *height = val[3];
  free(val);
}

/* ***************************************************************************
 *
 * *************************************************************************** */
unsigned char *NET_GetDesktopNames(unsigned int *size)
{
  return NET_AtomGet(Scr.Root, ATOM_NET_DESKTOP_NAMES, XA_UTF8_STRING, size);
}

/* ***************************************************************************
 *
 * *************************************************************************** */
void NET_GetWindowSystemTray(void)
{
  CARD32 *list;
  unsigned int list_size;
  int i,s;

  list = NET_AtomGet(Scr.Root, ATOM_KDE_NET_SYSTEM_TRAY_WINDOWS, XA_WINDOW,
		     &list_size);
  if (list == NULL) {
    fprintf(stderr,"******* SYS TRAY: NULL\n");
    return;
  }

  s = (list_size / (sizeof(CARD32)));
  fprintf(stderr,"******* SYS TRAY: ");
  for(i=0;i<s;i++)
    fprintf(stderr,"0x%lx ",list[i]);
  fprintf(stderr,"\n");
  free(list);
}

/* ***************************************************************************
 * 
 * *************************************************************************** */

/* ***************************************************************************
 * NET set functions
 * *************************************************************************** */

/* ***************************************************************************
 * set the list (for a task bar)
 * *************************************************************************** */
void NET_SetClientList()
{
  Window *wl = NULL;
  int i,nbr;
  NetWindow *t;
  
  nbr = 0;
  t = Start;

  while(t!= NULL) {
    nbr++;
    t = t->next;
  }

  if (nbr != 0)
  {
    wl = (Window *)safemalloc(sizeof (Window) * nbr);
    if (!wl)
    {
      fprintf (stderr, "NET_SetClientList: malloc failed!\n");
      return;
    }
    t = Start;
    i = 0;
    while (t)
    {
      if (!(DO_SKIP_WINDOW_LIST(t) && ForceSkipList))
	wl[i++] = t->w;
      t = t->next;
    }
    XChangeProperty(dpy, Scr.Root, ATOM_NET_CLIENT_LIST, XA_WINDOW,
		    32, PropModeReplace, (unsigned char *)wl, i);
    free (wl);
  }
  else
  {
    XChangeProperty(dpy, Scr.Root, ATOM_NET_CLIENT_LIST, XA_WINDOW,
		    32, PropModeReplace, (unsigned char *)NULL, 0);
  }
}

/* ***************************************************************************
 * set the staking list (for a pager)
 * *************************************************************************** */
void NET_SetStackingList(void)
{
  Window *wl = NULL;
  int i,nbr;
  Item *temp = (&Stacking)->head;
  NetWindow *t;
  
  nbr = 0;

  while(temp != NULL) {
    nbr++;
    temp = temp->next;
  }

  if (nbr != 0)
  {
    wl = (Window *)safemalloc(sizeof (Window) * nbr);
    if (!wl)
    {
      fprintf (stderr, "NET_SetStackingList: malloc failed!\n");
      return;
    }
    temp = (&Stacking)->head;
    i = 0;
    while (temp != NULL)
    {
      t = GetNetWindow(temp->w);
      if (!(ForcePagerList && DO_SKIP_PAGER_WINDOW_LIST(t)))
	wl[i++] = temp->w;
      temp = temp->next;
    }
    XChangeProperty(dpy, Scr.Root, ATOM_NET_CLIENT_LIST_STACKING, XA_WINDOW,
		    32, PropModeReplace, (unsigned char *)wl, i);
    free (wl);
  }
  else
  {
    XChangeProperty(dpy, Scr.Root, ATOM_NET_CLIENT_LIST_STACKING, XA_WINDOW,
		    32, PropModeReplace, (unsigned char *)NULL, 0);
  }
}

/* ***************************************************************************
 *
 * *************************************************************************** */
void NET_SetKdeSysTrayList(void)
{
  Window *wl = NULL;
  int i,nbr;
  Item *temp = (&KdeSysTray)->head;
  
  nbr = 0;

  while(temp != NULL) {
    nbr++;
    temp = temp->next;
  }

  if (nbr != 0)
  {
    wl = (Window *)safemalloc(sizeof (Window) * nbr);
    if (!wl)
    {
      fprintf (stderr, "NET_SetSystemTrayList: malloc failed!\n");
      return;
    }
    temp = (&KdeSysTray)->head;
    i = 0;
    while (temp != NULL)
    {
      wl[i] = temp->w;
      i++;
      temp = temp->next;
    }
    XChangeProperty(dpy, Scr.Root, ATOM_KDE_NET_SYSTEM_TRAY_WINDOWS, XA_WINDOW,
		    32, PropModeReplace, (unsigned char *)wl, nbr);
    free (wl);
  }
  else
  {
    XChangeProperty(dpy, Scr.Root, ATOM_KDE_NET_SYSTEM_TRAY_WINDOWS, XA_WINDOW,
		    32, PropModeReplace, (unsigned char *)NULL, 0);
  }
}

/* ***************************************************************************
 *
 * *************************************************************************** */
void NET_SetFrameStrut(NetWindow *netwin)
{
  CARD32 val[4];
  
  val[0] = netwin->frame_strut.left;
  val[1] = netwin->frame_strut.right;
  val[2] = netwin->frame_strut.top;
  val[3] = netwin->frame_strut.bottom;
  XChangeProperty(dpy, netwin->w, ATOM_KDE_NET_WM_FRAME_STRUT, XA_CARDINAL, 32,
		  PropModeReplace, (unsigned char *)&val, 4);
}

/* *************************************************************************** 
 * Desktop functions
 * *************************************************************************** */

void NET_SetCurrentDesktop(void)
{
  CARD32 val;

  val = Scr.CurrentDesk;

  if (val >= NumberOfDesktops && val <= MAX_NUMBER_OF_DESKTOPS)
  {
    NET_SetNumberOfDesktops();
  }

  if (val >= 0)
    XChangeProperty(dpy, Scr.Root, ATOM_NET_CURRENT_DESKTOP, XA_CARDINAL, 32,
		    PropModeReplace, (unsigned char *)&val, 1);
}

void NET_SetNumberOfDesktops(void)
{
  CARD32 val;

  if (Scr.CurrentDesk >= NumberOfDesktops && 
      Scr.CurrentDesk < MAX_NUMBER_OF_DESKTOPS)
    NumberOfDesktops = Scr.CurrentDesk + 1;

  val = (CARD32) NumberOfDesktops;
  XChangeProperty(dpy, Scr.Root, ATOM_NET_NUMBER_OF_DESKTOPS, XA_CARDINAL,
		  32, PropModeReplace, (unsigned char *)&val, 1);
}

void NET_SetDesktopGeometry(void)
{
  CARD32 val[2];

  val[0] = Scr.VWidth;
  val[1] = Scr.VHeight;

  XChangeProperty(dpy, Scr.Root, ATOM_NET_DESKTOP_GEOMETRY, XA_CARDINAL, 32,
		  PropModeReplace, (unsigned char *)&val, 2);
}

void NET_SetDesktopViewPort(void)
{
  CARD32 val[MAX_NUMBER_OF_DESKTOPS][2];
  int i = 0;

  while(i < NumberOfDesktops) {
    val[i][0] = Scr.Vx;
    val[i][1] = Scr.Vy;
    i++;
  }
  
  XChangeProperty(dpy, Scr.Root,ATOM_NET_DESKTOP_VIEWPORT , XA_CARDINAL,
		  32, PropModeReplace, (unsigned char *)&val, i*2);
}

void NET_SetDesktopWorkArea(void)
{
  CARD32 val[MAX_NUMBER_OF_DESKTOPS][4];
  int i = 0;
  int left,right,top,bottom;
  
  while(i < NumberOfDesktops) {
    NET_GetStrut(&left,&right,&top,&bottom,i,0);
    val[i][0] = left;
    val[i][1] = top;
    val[i][2] = Scr.MyDisplayWidth - (left + right);
    val[i][3] = Scr.MyDisplayHeight - (top + bottom);
    i++;
  }
  
  XChangeProperty(dpy, Scr.Root, ATOM_NET_WORKAREA, XA_CARDINAL, 32,
		  PropModeReplace, (unsigned char *)&val, i*4);
}

void NET_SetDesktopNames(unsigned char *val, int i)
{
  
  XChangeProperty(dpy, Scr.Root, ATOM_NET_DESKTOP_NAMES, XA_UTF8_STRING,
		    8, PropModeReplace, (unsigned char *)val, i);

}

/* ***************************************************************************
 * Visible (Icon) Name
 * *************************************************************************** */
void NET_SetVisibleName(NetWindow *netwin, Bool is_icon_name)
{
  unsigned char *val;
  unsigned char *visible_name = NULL;
  unsigned char *name;
  unsigned int size = 0;
  unsigned int s = 0;
  unsigned i,my_len;
  int j = 0;
  int *cw_count;
  unsigned count = 0;
  unsigned count0 = 0;
  unsigned count1 = 0;
  Bool done = False;
  NetWindow *t;
  Atom a_get, a_set;

  if (is_icon_name)
  {
    a_get = ATOM_NET_WM_ICON_NAME;
    a_set = ATOM_NET_WM_ICON_VISIBLE_NAME;
    cw_count = &netwin->net_wm_icon_visible_name_count;
  }
  else
  {
    a_get = ATOM_NET_WM_NAME;
    a_set = ATOM_NET_WM_VISIBLE_NAME;
    cw_count = &netwin->net_wm_visible_name_count;
  }

  val = NET_AtomGet(netwin->w, a_get, XA_UTF8_STRING, &size);

  if (val == NULL)
  {
    if (is_icon_name)
    {
      SET_HAS_NET_WM_ICON_NAME(netwin,False);
      if (netwin->net_wm_icon_name == NULL)
	return;
      name = netwin->net_wm_icon_name;
    }
    else
    {
      SET_HAS_NET_WM_NAME(netwin,False);
      if (netwin->net_wm_name == NULL)
	return;
      name = netwin->net_wm_name;
    }
  }
  else
  {
    if (val[size-1] != 0)
      s=size+1;
    else
      s=size;

    if (is_icon_name)
    {
      SET_HAS_NET_WM_ICON_NAME(netwin,True);
      if (netwin->net_wm_icon_name != NULL) {
	free(netwin->net_wm_icon_name);
	netwin->net_wm_icon_name = NULL;
      }
      netwin->net_wm_icon_name = 
	(unsigned char *)safemalloc(sizeof(unsigned char)*s);
      memcpy(netwin->net_wm_icon_name,val,size);
      if (s != size)
	netwin->net_wm_icon_name[s-1] = 0;
      name = netwin->net_wm_icon_name;
    }
    else
    {
      SET_HAS_NET_WM_NAME(netwin,True);
      if (netwin->net_wm_name != NULL) {
	free(netwin->net_wm_name);
	netwin->net_wm_name = NULL;
      }
      netwin->net_wm_name = 
	(unsigned char *)safemalloc(sizeof(unsigned char)*s);
      memcpy(netwin->net_wm_name,val,size);
      if (s != size)
	netwin->net_wm_name[s-1] = 0;
      name = netwin->net_wm_name;
    }

    free(val);
  }


  my_len = strlen(name);
  if (s == 0)
    s = my_len + 1;

  while(!done) {
    t = Start;
    count++;
    done = True;
    while (t != NULL && done) {
      char *tmp_name = (is_icon_name)? t->net_wm_icon_name : t->net_wm_name;
      int tmp_count = (is_icon_name)?
	t->net_wm_icon_visible_name_count : t->net_wm_visible_name_count;
      if (t->w != netwin->w && tmp_name != NULL) {
	if (strlen(tmp_name) == my_len && memcmp(name,tmp_name,my_len+1) == 0 &&
	    count == tmp_count)
	  done = False;
      }
      t = t->next;
    }
    if (count == 99)
      done = True;
  }

  *cw_count = count;

  size = s;
  if (count > 1 && count < 10) {
    count0 = 0x30 + count;
    count1 = 0;
    size = s + 4;
  }
  else if (count > 9 && count < 100) {
    count1 = (int)(count/10);
    count0 = 0x30 + (count - count1*10);
    count1 = 0x30 + count1;
    size = s + 5;
  }
  else if (count > 99) {
    count0 = 0x39;
    count1 = 0x39;
    size = s + 5;
  }

  visible_name = (unsigned char *)safemalloc(sizeof(unsigned char)*size);
  for(i=0;i<s-1;i++) {
    visible_name[j++] = name[i];
  }
  if (count0) {
    visible_name[j++] = 0x20;
    visible_name[j++] = 0x28;
    if (count1)
      visible_name[j++] = count1;
    visible_name[j++] = count0;
    visible_name[j++] = 0x29;
  }
  visible_name[j++] = 0x00;
  XChangeProperty(dpy, netwin->w, a_set, XA_UTF8_STRING, 8,
		  PropModeReplace, (unsigned char *)visible_name,
		  size);
  free(visible_name);
}

/* ***************************************************************************
 * 
 * *************************************************************************** */
void NET_DeleteVisibleNames(Window w)
{
  unsigned char *val1,*val2;
  unsigned int size = 0;

  val2 = NET_AtomGet(w, ATOM_NET_WM_VISIBLE_NAME, XA_UTF8_STRING, &size);
  if (val2 != NULL) {
    XDeleteProperty(dpy, w, ATOM_NET_WM_VISIBLE_NAME);
    free(val2);
  }

  val1 = NET_AtomGet(w, ATOM_NET_WM_ICON_VISIBLE_NAME, XA_UTF8_STRING, &size);
  if (val1 != NULL) {
    XDeleteProperty(dpy, w, ATOM_NET_WM_ICON_VISIBLE_NAME);
    free(val1);
  }

}

/* ***************************************************************************
 * The active window
 * *************************************************************************** */
void NET_SetActiveWindow(void)
{
  CARD32 w;

  if (FocusWin != NULL)
    w = FocusWin->w;
  else
    w = None;

  XChangeProperty(dpy, Scr.Root, ATOM_NET_ACTIVE_WINDOW, XA_WINDOW, 32,
		    PropModeReplace, (unsigned char *)&w, 1);
}

/* ***************************************************************************
 * the state of window
 * *************************************************************************** */
void NET_SetWMState(NetWindow *netwin) 
{
  Atom list[10];
  int i = 0;
  CARD32 desk = netwin->desk;
  CARD32 fvwm_sot = STAYS_ON_TOP_NONE;
  Bool pager_skip_list_set = False;
  Bool taskbar_skip_list_set = False;
  unsigned int s = 0;
  CARD32 *prev_fvwm_sot = 
    NET_AtomGet(netwin->w, ATOM_FVWM_NET_STAYS_ON_TOP, XA_CARDINAL, &s);

  if (prev_fvwm_sot != NULL) {
    fvwm_sot = prev_fvwm_sot[0];
    free(prev_fvwm_sot);
  }

  if (IS_STICKY(netwin)) {
    desk = 0xFFFFFFFE;
    list[i++] = ATOM_NET_WM_STATE_STICKY;
  }
  
  if (IS_MAXIMIZED(netwin)) {
    list[i++] = ATOM_NET_WM_STATE_MAXIMIZED_VERT;
    list[i++] = ATOM_NET_WM_STATE_MAXIMIZED_HORIZ;
    list[i++] = ATOM_NET_WM_STATE_MAXIMIZED_HORZ;
  }

  if (IS_SHADED(netwin))
    list[i++] = ATOM_NET_WM_STATE_SHADED;
  
  if (DO_SKIP_WINDOW_LIST(netwin)) {
    list[i++] = ATOM_NET_WM_STATE_SKIP_TASKBAR;
    taskbar_skip_list_set = True;
    if (PagerSkipListUseSkipList) {
      pager_skip_list_set = True;
      list[i++] = ATOM_NET_WM_STATE_SKIP_PAGER;
    }
  }

  if (DO_SKIP_PAGER_WINDOW_LIST(netwin)) {
    if (!pager_skip_list_set)
      list[i++] = ATOM_NET_WM_STATE_SKIP_PAGER;
    if (SkipListUsePagerSkipList && !taskbar_skip_list_set)
      list[i++] = ATOM_NET_WM_STATE_SKIP_TASKBAR;
  }

  if (netwin->layer >= StaysOnTopLayer) {
    list[i++] = ATOM_NET_WM_STATE_STAYS_ON_TOP;
    if (fvwm_sot ==  STAYS_ON_TOP_NONE)
      fvwm_sot = STAYS_ON_TOP_SET_BY_USER;
  }

  if (IS_NET_STAYS_ON_TOP(netwin))
    fvwm_sot = STAYS_ON_TOP_SET_BY_APPLICATION;

  if (IS_NET_MODAL(netwin)) {
    list[i++] = ATOM_NET_WM_STATE_MODAL;
  }

  if (i > 0)
    XChangeProperty(dpy, netwin->w, ATOM_NET_WM_STATE, XA_ATOM, 32,
		    PropModeReplace,(unsigned char *)list, i);
  else
    XDeleteProperty(dpy, netwin->w, ATOM_NET_WM_STATE);
  
  XChangeProperty(dpy, netwin->w, ATOM_NET_WM_DESKTOP, XA_CARDINAL, 32,
		  PropModeReplace,(unsigned char *)&desk, 1);

  XChangeProperty(dpy, netwin->w, ATOM_FVWM_NET_STAYS_ON_TOP, XA_CARDINAL, 32,
		  PropModeReplace,(unsigned char *)&fvwm_sot, 1);
}

/* ***************************************************************************
 * Set "All" functions
 * *************************************************************************** */
void NET_SetAllWMState(void)
{
  NetWindow *t;

  t= Start;
  while(t!= NULL) {
    NET_SetWMState(t);
    t = t->next;
  }
}

void NET_SetAll(void)
{
  NET_SetKdeSysTrayList();
  NET_SetClientList();
  NET_SetStackingList();
  NET_SetCurrentDesktop();
  NET_SetNumberOfDesktops();
  NET_SetDesktopGeometry();
  NET_SetDesktopViewPort();
  NET_SetDesktopWorkArea();
  NET_SetActiveWindow();
  NET_SetAllWMState();
}

/* ***************************************************************************
 * NET Initialization
 * *************************************************************************** */

void NET_Init(void)
{
  int i;
  Atom list[46];
  CARD32 val;
  XTextProperty text;
  CARD32 utf_name[5];
  Atom atom_set;
  char *names[1];

  /* init all the atoms */
  NET_AtomsInit();

  /* check for an other FvwmNetHints */
  CloseOldFvwmNetHints();

  /* create the net windows */
  Scr.Net_w = XCreateSimpleWindow (dpy, Scr.Root, -10, -10, 7, 7, 0, 0, 0);
  atom_set = ATOM_WM_DELETE_WINDOW;
  XSetWMProtocols(dpy,Scr.Net_w,&atom_set,1);

  names[0] = "FVWM";
  if (XStringListToTextProperty(names, 1, &text))
  {
    XSetWMName(dpy, Scr.Net_w, &text);
    XFree(text.value);
  }

  /* FVWM in UTF8 */
  utf_name[0] = 0x46;
  utf_name[1] = 0x56;
  utf_name[2] = 0x57;
  utf_name[3] = 0x4D;
  utf_name[4] = 0x0;

  XChangeProperty(dpy, Scr.Net_w, ATOM_NET_WM_NAME, XA_UTF8_STRING, 32,
		  PropModeReplace, (unsigned char *)&utf_name, 5);

  val = Scr.Net_w;
  XChangeProperty(dpy, Scr.Root, ATOM_NET_SUPPORTING_WM_CHECK, XA_WINDOW, 32,
		  PropModeReplace, (unsigned char *)&val, 1);
  XChangeProperty(dpy, Scr.Net_w, ATOM_NET_SUPPORTING_WM_CHECK, XA_WINDOW, 32,
		  PropModeReplace, (unsigned char *)&val, 1);

  val = getpid();
  XChangeProperty(dpy, Scr.Net_w, ATOM_NET_WM_PID, XA_CARDINAL, 32,
		  PropModeReplace, (unsigned char *)&val, 1);

  /* these indicate what NET compliance properties we support */
  i = 0;
  list[i++] = ATOM_NET_SUPPORTED;
  list[i++] = ATOM_NET_CLIENT_LIST;
  list[i++] = ATOM_NET_CLIENT_LIST_STACKING;
  list[i++] = ATOM_NET_NUMBER_OF_DESKTOPS;
  list[i++] = ATOM_NET_DESKTOP_GEOMETRY;
  list[i++] = ATOM_NET_DESKTOP_VIEWPORT;
  list[i++] = ATOM_NET_CURRENT_DESKTOP;
  list[i++] = ATOM_NET_DESKTOP_NAMES;
  list[i++] = ATOM_NET_ACTIVE_WINDOW;
  list[i++] = ATOM_NET_WORKAREA;
  list[i++] = ATOM_NET_SUPPORTING_WM_CHECK;
  list[i++] = ATOM_NET_VIRTUAL_ROOTS; /* NO? */

  list[i++] = ATOM_NET_CLOSE_WINDOW;
  list[i++] = ATOM_NET_WM_MOVERESIZE;

  list[i++] = ATOM_NET_WM_NAME;
  list[i++] = ATOM_NET_WM_VISIBLE_NAME;
  list[i++] = ATOM_NET_WM_ICON_NAME;
  list[i++] = ATOM_NET_WM_ICON_VISIBLE_NAME;
  list[i++] = ATOM_NET_WM_DESKTOP;
  list[i++] = ATOM_NET_WM_WINDOW_TYPE;
  list[i++] = ATOM_NET_WM_WINDOW_TYPE_DESKTOP;
  list[i++] = ATOM_NET_WM_WINDOW_TYPE_DOCK;
  list[i++] = ATOM_NET_WM_WINDOW_TYPE_TOOLBAR;
  list[i++] = ATOM_NET_WM_WINDOW_TYPE_MENU;
  list[i++] = ATOM_NET_WM_WINDOW_TYPE_DIALOG;
  list[i++] = ATOM_NET_WM_WINDOW_TYPE_NORMAL;
  list[i++] = ATOM_NET_WM_STATE;
  list[i++] = ATOM_NET_WM_STATE_MODAL;
  list[i++] = ATOM_NET_WM_STATE_STICKY;
  list[i++] = ATOM_NET_WM_STATE_MAXIMIZED_VERT;
  list[i++] = ATOM_NET_WM_STATE_MAXIMIZED_HORIZ;
  list[i++] = ATOM_NET_WM_STATE_MAXIMIZED_HORZ;
  list[i++] = ATOM_NET_WM_STATE_SHADED;
  list[i++] = ATOM_NET_WM_STATE_SKIP_TASKBAR;
  list[i++] = ATOM_NET_WM_STATE_SKIP_PAGER;
  list[i++] = ATOM_NET_WM_STATE_STAYS_ON_TOP;
  list[i++] = ATOM_NET_WM_STRUT;
  list[i++] = ATOM_NET_WM_ICON_GEOMETRY;
  list[i++] = ATOM_NET_WM_ICON;
  list[i++] = ATOM_NET_WM_PID;
  list[i++] = ATOM_NET_WM_HANDLED_ICON;
  list[i++] = ATOM_NET_WM_PING;
  list[i++] = ATOM_NET_WM_HANDLED_ICON;
  list[i++] = ATOM_KDE_NET_SYSTEM_TRAY_WINDOWS;
  list[i++] = ATOM_KDE_NET_WM_SYSTEM_TRAY_WINDOW_FOR;
  list[i++] = ATOM_KDE_NET_WM_FRAME_STRUT;

  XChangeProperty(dpy, Scr.Root, ATOM_NET_SUPPORTED, XA_ATOM, 32,
		  PropModeReplace,(unsigned char *)list, i);

  XSelectInput(dpy, Scr.Net_w,PropertyChangeMask);
}


/* ***************************************************************************
 * the state of window
 * *************************************************************************** */
void CloseOldFvwmNetHints(void)
{
  Window *win;
  unsigned char *val;
  Window w;
  int size;

  win = NET_AtomGet(Scr.Root, ATOM_NET_SUPPORTING_WM_CHECK, XA_WINDOW, &size);

  if (win == NULL) {
    return;
  }

  w = win[0];
  val = NET_AtomGet(w, ATOM_NET_WM_NAME, XA_UTF8_STRING, &size);
  
  if (val == NULL) {
    free(win);
    return;
  }

#ifdef NET_HINTS_WARNING_MSG
  fprintf(stderr,"%s: WARNING: a copy of FvwmNetHints already run! We shall \n\
              delete this copy, but if you want to restart FvwmNetHints\n\
              it is better to first kill it and then starting it again.\n",
	  MyName);
#endif
  free(val);
  send_clientmessage(dpy,w,ATOM_WM_DELETE_WINDOW,CurrentTime);
  /* 10 ms */
  usleep(10000);
}

/* ***************************************************************************
 * 
 * *************************************************************************** */
void NET_AtomsInit(void)
{
  
  /* Root Window properties (+ Related Message) */
  ATOM_NET_SUPPORTED            = XIA("_NET_SUPPORTED");
  ATOM_NET_CLIENT_LIST          = XIA("_NET_CLIENT_LIST");
  ATOM_NET_CLIENT_LIST_STACKING = XIA("_NET_CLIENT_LIST_STACKING");
  ATOM_NET_NUMBER_OF_DESKTOPS   = XIA("_NET_NUMBER_OF_DESKTOPS");
  ATOM_NET_DESKTOP_GEOMETRY     = XIA("_NET_DESKTOP_GEOMETRY");
  ATOM_NET_DESKTOP_VIEWPORT     = XIA("_NET_DESKTOP_VIEWPORT");
  ATOM_NET_CURRENT_DESKTOP      = XIA("_NET_CURRENT_DESKTOP");
  ATOM_NET_DESKTOP_NAMES        = XIA("_NET_DESKTOP_NAMES");
  ATOM_NET_ACTIVE_WINDOW        = XIA("_NET_ACTIVE_WINDOW");
  ATOM_NET_WORKAREA             = XIA("_NET_WORKAREA");
  ATOM_NET_SUPPORTING_WM_CHECK  = XIA("_NET_SUPPORTING_WM_CHECK");
  ATOM_NET_VIRTUAL_ROOTS        = XIA("_NET_VIRTUAL_ROOTS");

  /* Other Root window message */
  ATOM_NET_CLOSE_WINDOW  = XIA("_NET_CLOSE_WINDOW");
  ATOM_NET_WM_MOVERESIZE = XIA("_NET_WM_MOVERESIZE");

  /* Applications windows properties */
  ATOM_NET_WM_NAME                  = XIA("_NET_WM_NAME");
  ATOM_NET_WM_VISIBLE_NAME          = XIA("_NET_WM_VISIBLE_NAME");
  ATOM_NET_WM_ICON_NAME             = XIA("_NET_WM_ICON_NAME");
  ATOM_NET_WM_ICON_VISIBLE_NAME     = XIA("_NET_WM_ICON_VISIBLE_NAME");
  ATOM_NET_WM_DESKTOP               = XIA("_NET_WM_DESKTOP");
  
  ATOM_NET_WM_WINDOW_TYPE           = XIA("_NET_WM_WINDOW_TYPE");  
  ATOM_NET_WM_WINDOW_TYPE_DESKTOP   = XIA("_NET_WM_WINDOW_TYPE_DESKTOP");
  ATOM_NET_WM_WINDOW_TYPE_DOCK      = XIA("_NET_WM_WINDOW_TYPE_DOCK");
  ATOM_NET_WM_WINDOW_TYPE_TOOLBAR   = XIA("_NET_WM_WINDOW_TYPE_TOOLBAR");
  ATOM_NET_WM_WINDOW_TYPE_MENU      = XIA("_NET_WM_WINDOW_TYPE_MENU");
  ATOM_NET_WM_WINDOW_TYPE_DIALOG    = XIA("_NET_WM_WINDOW_TYPE_DIALOG");
  ATOM_NET_WM_WINDOW_TYPE_NORMAL    = XIA("_NET_WM_WINDOW_TYPE_NORMAL");
  
  ATOM_NET_WM_STATE                 = XIA("_NET_WM_STATE");  
  ATOM_NET_WM_STATE_MODAL           = XIA("_NET_WM_STATE_MODAL");  
  ATOM_NET_WM_STATE_STICKY          = XIA("_NET_WM_STATE_STICKY");  
  ATOM_NET_WM_STATE_MAXIMIZED_VERT  = XIA("_NET_WM_STATE_MAXIMIZED_VERT");
  /* HORIZ or HORZ ? */
  ATOM_NET_WM_STATE_MAXIMIZED_HORIZ = XIA("_NET_WM_STATE_MAXIMIZED_HORIZ");
  ATOM_NET_WM_STATE_MAXIMIZED_HORZ  = XIA("_NET_WM_STATE_MAXIMIZED_HORZ");
  ATOM_NET_WM_STATE_SHADED          = XIA("_NET_WM_STATE_SHADED");  
  ATOM_NET_WM_STATE_SKIP_TASKBAR    = XIA("_NET_WM_STATE_SKIP_TASKBAR");
  ATOM_NET_WM_STATE_SKIP_PAGER      = XIA("_NET_WM_STATE_SKIP_PAGER");
  /* the next one is not in the spec but KDE use it */
  ATOM_NET_WM_STATE_STAYS_ON_TOP    = XIA("_NET_WM_STATE_STAYS_ON_TOP");
 
  ATOM_NET_WM_STRUT                 = XIA("_NET_WM_STRUT");
  ATOM_NET_WM_ICON_GEOMETRY         = XIA("_NET_WM_ICON_GEOMETRY");
  ATOM_NET_WM_ICON                  = XIA("_NET_WM_ICON");
  ATOM_NET_WM_PID                   = XIA("_NET_WM_PID");
  ATOM_NET_WM_HANDLED_ICON          = XIA("_NET_WM_HANDLED_ICON");
  
  /* Window manager protocol */
  ATOM_NET_WM_PING      = XIA("_NET_WM_PING");
  ATOM_WM_DELETE_WINDOW = XIA("WM_DELETE_WINDOW");
  ATOM_WM_PROTOCOLS     = XIA("WM_PROTOCOLS");

  ATOM_WM_STATE = XIA("WM_STATE");

  /* KDE specific atoms: */
  ATOM_KDE_NET_SYSTEM_TRAY_WINDOWS  = XIA("_KDE_NET_SYSTEM_TRAY_WINDOWS");
  ATOM_KDE_NET_WM_SYSTEM_TRAY_WINDOW_FOR = 
    XIA("_KDE_NET_WM_SYSTEM_TRAY_WINDOW_FOR");
  ATOM_KDE_NET_WM_FRAME_STRUT = XIA("_KDE_NET_WM_FRAME_STRUT");

  /* Needed for the "name" atom */
  XA_UTF8_STRING = XIA("UTF8_STRING");

  /* FvwmNetHints atoms */
  ATOM_FVWM_NET_SET_MINI_ICON = XIA("_FVWM_NET_SET_MINI_ICON");
  ATOM_FVWM_NET_SET_ICON      = XIA("_FVWM_NET_SET_ICON");
  ATOM_FVWM_NET_STAYS_ON_TOP = XIA("_FVWM_NET_STAYS_ON_TOP");
}

/* ***************************************************************************
 * Motif Stuff to force no decorations
 * *************************************************************************** */
int SetUpMwmHints(Window w, net_style style, CARD32 functions)
{
  PropMwmHints *old_prop;
  PropMwmHints prop;
  unsigned int size = 0;
  Atom ATOM_MOTIF_WM_HINTS = XIA("_MOTIF_WM_HINTS");

  old_prop = (PropMwmHints *)NET_AtomGet(w, ATOM_MOTIF_WM_HINTS,
					 ATOM_MOTIF_WM_HINTS, &size);
  if (old_prop != NULL && !style.force_decoration) {
    free(old_prop);
    return 0;
  }

  prop.flags = MWM_HINTS_DECORATIONS | MWM_HINTS_FUNCTIONS;
  prop.decorations = 0;
  prop.functions = functions;
  if (old_prop != NULL) {
    prop.flags = old_prop->flags;
    prop.decorations= old_prop->decorations;
    prop.functions = old_prop->functions;
    if (prop.decorations & MWM_DECOR_BORDER)
      prop.decorations &= ~MWM_DECOR_BORDER;
    if (prop.decorations & MWM_DECOR_TITLE)
      prop.decorations &= ~MWM_DECOR_TITLE;
    prop.functions = old_prop->functions;
    if (!(old_prop->flags & MWM_HINTS_DECORATIONS))
      prop.flags |= MWM_HINTS_DECORATIONS;
    free(old_prop);
  }
  if (style.has_border)
    prop.decorations |= MWM_DECOR_BORDER;
  if (style.has_title)
    prop.decorations |= MWM_DECOR_TITLE;

   XChangeProperty(dpy, w, ATOM_MOTIF_WM_HINTS, ATOM_MOTIF_WM_HINTS, 32,
		  PropModeReplace,(unsigned char *)&prop, 4);
   return 1;
}

/* ***************************************************************************
 * To force Never Never Focus
 * *************************************************************************** */
void SetUpWMHintsFocus(Window w, net_style style)
{
  XWMHints *wmhints = NULL;
  Bool set = False;

  wmhints = XGetWMHints(dpy, w);
  if (wmhints != NULL) {
    wmhints->flags |= InputHint;
    if ((style.never_focus && wmhints->input) || !(wmhints->flags & InputHint))
    {
      wmhints->input = False;
      set = True;
    }
    else if ((!style.never_focus && !wmhints->input) || 
	     !(wmhints->flags & InputHint))
    {
      set = True;
      wmhints->input = True;
    }
    if (set)
      XSetWMHints(dpy, w, wmhints);
    XFree(wmhints);
  }
}
