/* Copyright 2001, Olivier Chapuis. No guarantees or warantees or anything
 * are provided or implied in any way whatsoever. Use this program at your
 * own risk. Permission to use this program for any purpose is given,
 * as long as the copyright is kept intact. */

/* This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>

#include <libs/fvwmlib.h>
#include "libs/Module.h"
#include "fvwm/fvwm.h"

#include "FvwmNetHints.h"
#include "net.h"
#include "parse.h"
#include "net_debug.h"
#include "my_imlib.h"
#include "utf8.h"

extern char *MyName;
extern int fd[2];

extern ScreenInfo Scr;
extern Display *dpy;
extern NetWindow *Start;

extern int StaysOnTopLayer;
extern int StaysPutLayer;
#if HAVE_ICONV
extern char *IconvCharset;
#endif


void UpdateStaysOnTop(Bool accepted)
{
  NetWindow *t;
  char command[40];

  t = Start;

  while (t != NULL)
  {
    if(IS_NET_STAYS_ON_TOP(t)) {
      if (accepted && t->layer != StaysOnTopLayer) {
	sprintf(command,"Silent Layer 0 %i",StaysOnTopLayer);
	SendText(fd,command,t->w);
      }
      if (!accepted && t->layer != StaysPutLayer) {
	sprintf(command,"Silent Layer 0 %i",StaysPutLayer);
	SendText(fd,command,t->w);
      }
    }
    t = t->next;
  }

}

void UpdateMiniIconSize()
{
#ifdef HAVE_AN_IMLIB
  NetWindow *t;

  t = Start;

  while (t != NULL)
  {
    if (t->Imlib_mini != NULL) {
      BuildMiniIcon(t, NULL, True);
    }
    t = t->next;
  }
  SendMiniIcon(0);
#endif
}

void UpdateNetStyle(Atom a, net_style st)
{
  NetWindow *t;

  t = Start;

  while (t != NULL)
  {
    if (t->net_window_type == a) {
      NET_ReconfigureWindowType(t);
      SetUpWMHintsFocus(t->w, st);
    }
    t = t->next;
  }
}

void UpdateModalNetStyle(net_style st)
{
  NetWindow *t;

  t = Start;

  while (t != NULL)
  {
    if (IS_NET_MODAL(t)) {
      NET_ReconfigureWindowType(t);
      SetUpWMHintsFocus(t->w, st);
    }
    t = t->next;
  }
}

#if HAVE_ICONV
void UpdateVisibleNames(Bool is_delete)
{
  NetWindow *t;

  t = Start;
  while (t != NULL)
  {
    t->net_wm_visible_name_count = -1;
    t->net_wm_icon_visible_name_count = -1;
    if (is_delete)
      NET_DeleteVisibleNames(t->w);
    t = t->next;
  }

  if (is_delete)
    return;

  t = Start;
  while (t != NULL)
  {
    if (DO_HAS_NET_WM_NAME(t))
    {
      NET_SetVisibleName(t,False);
    }
    else if (t->window_name != NULL)
    {
      if (t->net_wm_name != NULL) {
	free(t->net_wm_name);
	t->net_wm_name = NULL;
      }
      t->net_wm_name = 
	(unsigned char *)UTF8_charset2utf8(IconvCharset,t->window_name);
      NET_SetVisibleName(t,False);
    }

    if (DO_HAS_NET_WM_ICON_NAME(t))
    {
	NET_SetVisibleName(t,True);
    }
    else if (t->icon_name != NULL)
    {
      if (t->net_wm_icon_name != NULL) {
	free(t->net_wm_icon_name);
	t->net_wm_icon_name = NULL;
      }
      t->net_wm_icon_name = 
	(unsigned char *)UTF8_charset2utf8(IconvCharset,t->icon_name);
      NET_SetVisibleName(t,True);
    }

    t = t->next;
  }
}
#endif
