/*
 * Copyright 2001, Olivier Chapuis. No guarantees or warantees or anything
 * are provided or implied in any way whatsoever. Use this program at your
 * own risk. Permission to use this program for any purpose is given,
 * as long as the copyright is kept intact. */

/* This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>

#include "Mallocs.h"
#include "libs/Module.h"

#include "FvwmNetHints.h"
#include "windowlist.h"
#include "net.h"

extern char *MyName;

/* ***************************************************************************
 * InitList - Initialize a list
 * *************************************************************************** */
void InitList(WindowList *list)
{
  list->head=list->tail=NULL;
}

void InitPingList(PingWindowList *list)
{
  list->head=list->tail=NULL;
}

/* ***************************************************************************
 * AddItem - Allocates spaces for and appends an item to the list
 * *************************************************************************** */
void AddItem(WindowList *list, Window w)
{
  Item *new;
  new=(Item *)safemalloc(sizeof(Item));
  new->w=w;
  new->next=NULL;

  if (list->tail==NULL) list->head=list->tail=new;
  else {
    list->tail->next=new;
    list->tail=new;
  }
}

void AddPingItem(PingWindowList *list, Window w,
		 int id,int action, time_t time, time_t time_out, pid_t pid)
{
  PingItem *new;
  new=(PingItem *)safemalloc(sizeof(PingItem));
  new->w=w;
  new->p_id = id;
  new->p_action = action;
  new->p_time = time;
  new->p_time_out = time_out;
  new->p_pid = pid;
  new->next=NULL;

  if (list->tail==NULL) list->head=list->tail=new;
  else {
    list->tail->next=new;
    list->tail=new;
  }
}

/* ***************************************************************************
 * return 1 if the Window is in the list, 0 if not
 * *************************************************************************** */
Bool IsItemInList(WindowList *list,Window w)
{
  Bool r = False;

  Item *temp = list->head;
  while(temp != NULL && !r)
  {
    if (temp->w == w) {
      r = True;
    }
    temp = temp->next;
  }
  return r;
}

/* ***************************************************************************
 * DeleteItem - Deletes an item from the list
 * *************************************************************************** */
void DeleteItem(WindowList *list,Window w)
{
  Item *temp,*temp2;
  int i;

  if (list->head==NULL) return;
  if (list->head->w==w)
  {
    temp2=list->head;
    temp=list->head=list->head->next;
  }
  else
  {
    for(i=0,temp=list->head;temp->next!=NULL && temp->next->w!=w;
      i++,temp=temp->next);
    if (temp->next==NULL) return;
    temp2=temp->next;
    temp->next=temp2->next;
  }
  if (temp2==list->tail) list->tail=temp;
}

void DeletePingItem(PingWindowList *list,int id)
{
  PingItem *temp,*temp2;
  int i;

  if (list->head==NULL) return;
  if (list->head->p_id==id)
  {
    temp2=list->head;
    temp=list->head=list->head->next;
  }
  else
  {
    for(i=0,temp=list->head;temp->next!=NULL && temp->next->p_id!=id;
      i++,temp=temp->next);
    if (temp->next==NULL) return;
    temp2=temp->next;
    temp->next=temp2->next;
  }
  if (temp2==list->tail) list->tail=temp;
}


/******************************************************************************
  FreeList - Free the entire list of Items
******************************************************************************/
void FreeList(WindowList *list)
{
  Item *temp,*temp2;

  for(temp=list->head;temp!=NULL;)
  {
    temp2=temp;
    temp=temp->next;
    free(temp2);
  }
}

/* ***************************************************************************
 * IsPingItemInfo - says if the item is an info ping item
 * *************************************************************************** */
Bool IsPingItemInfo(PingWindowList *list, int id)
{
  Bool r = False;

  PingItem *temp = list->head;
  while(temp != NULL && !r)
  {
    if (temp->p_action == PING_ACTION_INFO && temp->p_id == id) {
      r = True;
    }
    temp = temp->next;
  }
  return r;
}
    
/* *****************************************************************************
 * RestackList - Restack the window list
 * ****************************************************************************/
void RestackList(WindowList *list, unsigned long *body, unsigned long length)
{

  Item *temp;
  Window w;
  Window *wins = NULL;
  int i,j;

  wins = (Window *) safemalloc (length * sizeof (Window));
  
  j = 0;
  for (i = 0; i < (length - FvwmPacketHeaderSize); i += 3)
  {
    w = body[i];
    temp = list->head;
    while((temp!= NULL)&&(temp->w != w))
      temp = temp->next;
    if (temp != NULL)
      wins[j++] = temp->w;
  }


  /* stuipd algo which work fine */
  for(i=0; i < j; i++) {
    DeleteItem(&*list, wins[i]);
  }

  for(i=j-1; i >= 0; i--) {
    AddItem(&*list,wins[i]);
  }

  free (wins);
}

